#################################################################################
# The Institute for the Design of Advanced Energy Systems Integrated Platform
# Framework (IDAES IP) was produced under the DOE Institute for the
# Design of Advanced Energy Systems (IDAES).
#
# Copyright (c) 2018-2024 by the software owners: The Regents of the
# University of California, through Lawrence Berkeley National Laboratory,
# National Technology & Engineering Solutions of Sandia, LLC, Carnegie Mellon
# University, West Virginia University Research Corporation, et al.
# All rights reserved.  Please see the files COPYRIGHT.md and LICENSE.md
# for full copyright and license information.
#################################################################################
"""
Test code: power plant costing of subaccounts 6.1. 
Subaccounts 6.1-6.8 consist of costing correlations for MEA solvent-based
carbon capture system, Reference: 
Kangkang Li, Ashleigh Cousins, Hai You, Paul Feron,
Weilang Luo, Jian Chen (2016). Systematic study of aqueous monoethanolamine
(MEA)-based CO2 capture process: Techno-economic assessment of the MEA
process and its improvements. Applied Energy, 165, 648-659.
"""

import os
import json
import pytest
from idaes.models_extra.power_generation.costing.power_plant_capcost import (
    QGESSCosting,
    QGESSCostingData,
)

from idaes.core.util.model_statistics import degrees_of_freedom
import pyomo.environ as pyo
from pyomo.environ import units as pyunits
from idaes.core import FlowsheetBlock
from idaes.core.solvers import get_solver
from pyomo.util.check_units import assert_units_consistent

from pyomo.common.fileutils import this_file_dir
from pyomo.common.tempfiles import TempfileManager
from idaes.models_extra.power_generation.costing.generic_ccs_capcost_custom_dict import (
    load_generic_ccs_costing_dictionary,
)

directory = this_file_dir()

# Get default solver for testing
solver = get_solver()


@pytest.mark.component
def test_regenerate_ccs_dictionaries():
    # Switch to a temporary directory to test regenerating the dictionaries
    _temp_context = TempfileManager.new_context()
    wrkdir = _temp_context.create_tempdir()

    generic_ccs_costing_params = load_generic_ccs_costing_dictionary(path=wrkdir)

    # check that expected file was generated
    dictionaries_path = os.path.join(wrkdir, "generic_ccs_costing_data.json")
    assert os.path.exists(dictionaries_path)

    # expected output of dictionary
    dictionaries_expected_output = {
        "6": {
            "B": {
                "5.1.a.epri": {
                    "Account Name": "Cansolv CO2 Removal System",
                    "BEC": 224191.4,
                    "BEC_units": "K$2018",
                    "Eng Fee": 0.2,
                    "Exponent": 2.788,
                    "Process Contingency": 0.18,
                    "Process Parameter": "CO2 Flowrate",
                    "Project Contingency": 0.2,
                    "RP Value": 493587.88,
                    "Units": "lb/hr",
                },
                "6.1.ccs": {
                    "Account Name": "MEA solvent capture system absorber",
                    "BEC": 8376.0,
                    "BEC_units": "K$2013",
                    "Eng Fee": 0.2,
                    "Exponent": 0.6,
                    "Process Contingency": 0.12,
                    "Process Parameter": "Absorber volume",
                    "Project Contingency": 0.2,
                    "RP Value": 1074.424688,
                    "Units": "m**3",
                },
                "6.2.ccs": {
                    "Account Name": "MEA solvent capture system absorber packing",
                    "BEC": 6117.0,
                    "BEC_units": "K$2013",
                    "Eng Fee": 0.2,
                    "Exponent": 0.6,
                    "Process Contingency": 0.12,
                    "Process Parameter": "Absorber packing volume",
                    "Project Contingency": 0.2,
                    "RP Value": 791.6813487,
                    "Units": "m**3",
                },
                "6.3.ccs": {
                    "Account Name": "MEA solvent capture system stripper",
                    "BEC": 2273.0,
                    "BEC_units": "K$2013",
                    "Eng Fee": 0.2,
                    "Exponent": 0.6,
                    "Process Contingency": 0.12,
                    "Process Parameter": "Stripper volume",
                    "Project Contingency": 0.2,
                    "RP Value": 419.6971436,
                    "Units": "m**3",
                },
                "6.4.ccs": {
                    "Account Name": "MEA solvent capture system stripper packing",
                    "BEC": 1464.0,
                    "BEC_units": "K$2013",
                    "Eng Fee": 0.2,
                    "Exponent": 0.6,
                    "Process Contingency": 0.12,
                    "Process Parameter": "Stripper packing volume",
                    "Project Contingency": 0.2,
                    "RP Value": 309.2505268,
                    "Units": "m**3",
                },
                "6.5.ccs": {
                    "Account Name": "MEA solvent capture system stripper condenser",
                    "BEC": 305.0,
                    "BEC_units": "K$2013",
                    "Eng Fee": 0.2,
                    "Exponent": 0.6,
                    "Process Contingency": 0.12,
                    "Process Parameter": "Stripper condenser area",
                    "Project Contingency": 0.2,
                    "RP Value": 800,
                    "Units": "m**2",
                },
                "6.6.ccs": {
                    "Account Name": "MEA solvent capture system stripper reboiler",
                    "BEC": 3183.0,
                    "BEC_units": "K$2013",
                    "Eng Fee": 0.2,
                    "Exponent": 0.6,
                    "Process Contingency": 0.12,
                    "Process Parameter": "Stripper reboiler area",
                    "Project Contingency": 0.2,
                    "RP Value": 4250,
                    "Units": "m**2",
                },
                "6.7.ccs": {
                    "Account Name": "MEA solvent capture system lean rich heat exchanger",
                    "BEC": 1502.0,
                    "BEC_units": "K$2013",
                    "Eng Fee": 0.2,
                    "Exponent": 0.6,
                    "Process Contingency": 0.12,
                    "Process Parameter": "Lean rich heat exchanger area",
                    "Project Contingency": 0.2,
                    "RP Value": 9100,
                    "Units": "m**2",
                },
                "6.8.ccs": {
                    "Account Name": "MEA solvent capture system lean solvent cooler",
                    "BEC": 522.0,
                    "BEC_units": "K$2013",
                    "Eng Fee": 0.2,
                    "Exponent": 0.6,
                    "Process Contingency": 0.12,
                    "Process Parameter": "Lean solvent cooler area",
                    "Project Contingency": 0.2,
                    "RP Value": 1200,
                    "Units": "m**2",
                },
                "6.9.1.ccs": {
                    "Account Name": "MEA solvent capture system flue gas blower cost component 1",
                    "BEC": 487.2,
                    "BEC_units": "K$2013",
                    "Eng Fee": 0.2,
                    "Exponent": 0.6,
                    "Process Contingency": 0.12,
                    "Process Parameter": "CO2 product flow rate",
                    "Project Contingency": 0.2,
                    "RP Value": 262349.78,
                    "Units": "lb/hr",
                },
                "6.9.2.ccs": {
                    "Account Name": "MEA solvent capture system flue gas blower cost component 2",
                    "BEC": 324.8,
                    "BEC_units": "K$2013",
                    "Eng Fee": 0.2,
                    "Exponent": 0.6,
                    "Process Contingency": 0.12,
                    "Process Parameter": "Flue gas inlet to absorber",
                    "Project Contingency": 0.2,
                    "RP Value": 672504.35,
                    "Units": "m**3/hr",
                },
                "6.10.1.ccs": {
                    "Account Name": "MEA solvent capture system flue gas direct contact cooler cost component 1",
                    "BEC": 1725.0,
                    "BEC_units": "K$2013",
                    "Eng Fee": 0.2,
                    "Exponent": 0.6,
                    "Process Contingency": 0.12,
                    "Process Parameter": "CO2 product flow rate",
                    "Project Contingency": 0.2,
                    "RP Value": 262349.78,
                    "Units": "lb/hr",
                },
                "6.10.2.ccs": {
                    "Account Name": "MEA solvent capture system flue gas direct contact cooler cost component 2",
                    "BEC": 1150.0,
                    "BEC_units": "K$2013",
                    "Eng Fee": 0.2,
                    "Exponent": 0.6,
                    "Process Contingency": 0.12,
                    "Process Parameter": "Flue gas inlet to absorber",
                    "Project Contingency": 0.2,
                    "RP Value": 672504.35,
                    "Units": "m**3/hr",
                },
                "6.11.1.ccs": {
                    "Account Name": "MEA solvent capture system flue gas direct contact cooler packing cost component 1",
                    "BEC": 1129.8,
                    "BEC_units": "K$2013",
                    "Eng Fee": 0.2,
                    "Exponent": 0.6,
                    "Process Contingency": 0.12,
                    "Process Parameter": "CO2 product flow rate",
                    "Project Contingency": 0.2,
                    "RP Value": 262349.78,
                    "Units": "lb/hr",
                },
                "6.11.2.ccs": {
                    "Account Name": "MEA solvent capture system flue gas direct contact cooler packing cost component 2",
                    "BEC": 753.2,
                    "BEC_units": "K$2013",
                    "Eng Fee": 0.2,
                    "Exponent": 0.6,
                    "Process Contingency": 0.12,
                    "Process Parameter": "Flue gas inlet to absorber",
                    "Project Contingency": 0.2,
                    "RP Value": 672504.35,
                    "Units": "m**3/hr",
                },
                "6.12.1.ccs": {
                    "Account Name": "MEA solvent capture system pretreatment pump cost component 1",
                    "BEC": 64.8,
                    "BEC_units": "K$2013",
                    "Eng Fee": 0.2,
                    "Exponent": 0.6,
                    "Process Contingency": 0.12,
                    "Process Parameter": "CO2 product flow rate",
                    "Project Contingency": 0.2,
                    "RP Value": 262349.78,
                    "Units": "lb/hr",
                },
                "6.12.2.ccs": {
                    "Account Name": "MEA solvent capture system pretreatment pump cost component 2",
                    "BEC": 43.2,
                    "BEC_units": "K$2013",
                    "Eng Fee": 0.2,
                    "Exponent": 0.6,
                    "Process Contingency": 0.12,
                    "Process Parameter": "Flue gas inlet to absorber",
                    "Project Contingency": 0.2,
                    "RP Value": 672504.35,
                    "Units": "m**3/hr",
                },
                "6.13.1.ccs": {
                    "Account Name": "MEA solvent capture system pretreatment cooler cost component 1",
                    "BEC": 121.2,
                    "BEC_units": "K$2013",
                    "Eng Fee": 0.2,
                    "Exponent": 0.6,
                    "Process Contingency": 0.12,
                    "Process Parameter": "CO2 product flow rate",
                    "Project Contingency": 0.2,
                    "RP Value": 262349.78,
                    "Units": "lb/hr",
                },
                "6.13.2.ccs": {
                    "Account Name": "MEA solvent capture system pretreatment cooler cost component 2",
                    "BEC": 80.8,
                    "BEC_units": "K$2013",
                    "Eng Fee": 0.2,
                    "Exponent": 0.6,
                    "Process Contingency": 0.12,
                    "Process Parameter": "Flue gas inlet to absorber",
                    "Project Contingency": 0.2,
                    "RP Value": 672504.35,
                    "Units": "m**3/hr",
                },
                "6.14.1.ccs": {
                    "Account Name": "MEA solvent capture system pretreatment tank cost component 1",
                    "BEC": 55.8,
                    "BEC_units": "K$2013",
                    "Eng Fee": 0.2,
                    "Exponent": 0.6,
                    "Process Contingency": 0.12,
                    "Process Parameter": "CO2 product flow rate",
                    "Project Contingency": 0.2,
                    "RP Value": 262349.78,
                    "Units": "lb/hr",
                },
                "6.14.2.ccs": {
                    "Account Name": "MEA solvent capture system pretreatment tank cost component 2",
                    "BEC": 37.2,
                    "BEC_units": "K$2013",
                    "Eng Fee": 0.2,
                    "Exponent": 0.6,
                    "Process Contingency": 0.12,
                    "Process Parameter": "Flue gas inlet to absorber",
                    "Project Contingency": 0.2,
                    "RP Value": 672504.35,
                    "Units": "m**3/hr",
                },
                "6.15.1.ccs": {
                    "Account Name": "MEA solvent capture system washing column cost component 1",
                    "BEC": 1705.2,
                    "BEC_units": "K$2013",
                    "Eng Fee": 0.2,
                    "Exponent": 0.6,
                    "Process Contingency": 0.12,
                    "Process Parameter": "CO2 product flow rate",
                    "Project Contingency": 0.2,
                    "RP Value": 262349.78,
                    "Units": "lb/hr",
                },
                "6.15.2.ccs": {
                    "Account Name": "MEA solvent capture system washing column cost component 2",
                    "BEC": 1136.8,
                    "BEC_units": "K$2013",
                    "Eng Fee": 0.2,
                    "Exponent": 0.6,
                    "Process Contingency": 0.12,
                    "Process Parameter": "Flue gas inlet to absorber",
                    "Project Contingency": 0.2,
                    "RP Value": 672504.35,
                    "Units": "m**3/hr",
                },
                "6.16.1.ccs": {
                    "Account Name": "MEA solvent capture system washing column packing cost component 1",
                    "BEC": 1236.0,
                    "BEC_units": "K$2013",
                    "Eng Fee": 0.2,
                    "Exponent": 0.6,
                    "Process Contingency": 0.12,
                    "Process Parameter": "CO2 product flow rate",
                    "Project Contingency": 0.2,
                    "RP Value": 262349.78,
                    "Units": "lb/hr",
                },
                "6.16.2.ccs": {
                    "Account Name": "MEA solvent capture system washing column packing cost component 2",
                    "BEC": 824.0,
                    "BEC_units": "K$2013",
                    "Eng Fee": 0.2,
                    "Exponent": 0.6,
                    "Process Contingency": 0.12,
                    "Process Parameter": "Flue gas inlet to absorber",
                    "Project Contingency": 0.2,
                    "RP Value": 672504.35,
                    "Units": "m**3/hr",
                },
                "6.17.1.ccs": {
                    "Account Name": "MEA solvent capture system washing solvent cooler cost component 1",
                    "BEC": 41.4,
                    "BEC_units": "K$2013",
                    "Eng Fee": 0.2,
                    "Exponent": 0.6,
                    "Process Contingency": 0.12,
                    "Process Parameter": "CO2 product flow rate",
                    "Project Contingency": 0.2,
                    "RP Value": 262349.78,
                    "Units": "lb/hr",
                },
                "6.17.2.ccs": {
                    "Account Name": "MEA solvent capture system washing solvent cooler cost component 2",
                    "BEC": 27.6,
                    "BEC_units": "K$2013",
                    "Eng Fee": 0.2,
                    "Exponent": 0.6,
                    "Process Contingency": 0.12,
                    "Process Parameter": "Flue gas inlet to absorber",
                    "Project Contingency": 0.2,
                    "RP Value": 672504.35,
                    "Units": "m**3/hr",
                },
                "6.18.1.ccs": {
                    "Account Name": "MEA solvent capture system washing solvent pump cost component 1",
                    "BEC": 10.2,
                    "BEC_units": "K$2013",
                    "Eng Fee": 0.2,
                    "Exponent": 0.6,
                    "Process Contingency": 0.12,
                    "Process Parameter": "CO2 product flow rate",
                    "Project Contingency": 0.2,
                    "RP Value": 262349.78,
                    "Units": "lb/hr",
                },
                "6.18.2.ccs": {
                    "Account Name": "MEA solvent capture system washing solvent pump cost component 2",
                    "BEC": 6.8,
                    "BEC_units": "K$2013",
                    "Eng Fee": 0.2,
                    "Exponent": 0.6,
                    "Process Contingency": 0.12,
                    "Process Parameter": "Flue gas inlet to absorber",
                    "Project Contingency": 0.2,
                    "RP Value": 672504.35,
                    "Units": "m**3/hr",
                },
                "6.19.1.ccs": {
                    "Account Name": "MEA solvent capture system condenser pump cost component 1",
                    "BEC": 21.0,
                    "BEC_units": "K$2013",
                    "Eng Fee": 0.2,
                    "Exponent": 0.6,
                    "Process Contingency": 0.12,
                    "Process Parameter": "CO2 product flow rate",
                    "Project Contingency": 0.2,
                    "RP Value": 262349.78,
                    "Units": "lb/hr",
                },
                "6.19.2.ccs": {
                    "Account Name": "MEA solvent capture system condenser pump cost component 2",
                    "BEC": 14.0,
                    "BEC_units": "K$2013",
                    "Eng Fee": 0.2,
                    "Exponent": 0.6,
                    "Process Contingency": 0.12,
                    "Process Parameter": "Flue gas inlet to absorber",
                    "Project Contingency": 0.2,
                    "RP Value": 672504.35,
                    "Units": "m**3/hr",
                },
                "6.20.1.ccs": {
                    "Account Name": "MEA solvent capture system stripper reflux drum cost component 1",
                    "BEC": 27.6,
                    "BEC_units": "K$2013",
                    "Eng Fee": 0.2,
                    "Exponent": 0.6,
                    "Process Contingency": 0.12,
                    "Process Parameter": "CO2 product flow rate",
                    "Project Contingency": 0.2,
                    "RP Value": 262349.78,
                    "Units": "lb/hr",
                },
                "6.20.2.ccs": {
                    "Account Name": "MEA solvent capture system stripper reflux drum cost component 2",
                    "BEC": 18.4,
                    "BEC_units": "K$2013",
                    "Eng Fee": 0.2,
                    "Exponent": 0.6,
                    "Process Contingency": 0.12,
                    "Process Parameter": "Flue gas inlet to absorber",
                    "Project Contingency": 0.2,
                    "RP Value": 672504.35,
                    "Units": "m**3/hr",
                },
                "6.21.1.ccs": {
                    "Account Name": "MEA solvent capture system lean solvent pump cost component 1",
                    "BEC": 95.4,
                    "BEC_units": "K$2013",
                    "Eng Fee": 0.2,
                    "Exponent": 0.6,
                    "Process Contingency": 0.12,
                    "Process Parameter": "CO2 product flow rate",
                    "Project Contingency": 0.2,
                    "RP Value": 262349.78,
                    "Units": "lb/hr",
                },
                "6.21.2.ccs": {
                    "Account Name": "MEA solvent capture system lean solvent pump cost component 2",
                    "BEC": 63.6,
                    "BEC_units": "K$2013",
                    "Eng Fee": 0.2,
                    "Exponent": 0.6,
                    "Process Contingency": 0.12,
                    "Process Parameter": "Flue gas inlet to absorber",
                    "Project Contingency": 0.2,
                    "RP Value": 672504.35,
                    "Units": "m**3/hr",
                },
                "6.22.1.ccs": {
                    "Account Name": "MEA solvent capture system solvent storage tank cost component 1",
                    "BEC": 205.2,
                    "BEC_units": "K$2013",
                    "Eng Fee": 0.2,
                    "Exponent": 0.6,
                    "Process Contingency": 0.12,
                    "Process Parameter": "CO2 product flow rate",
                    "Project Contingency": 0.2,
                    "RP Value": 262349.78,
                    "Units": "lb/hr",
                },
                "6.22.2.ccs": {
                    "Account Name": "MEA solvent capture system solvent storage tank cost component 2",
                    "BEC": 136.8,
                    "BEC_units": "K$2013",
                    "Eng Fee": 0.2,
                    "Exponent": 0.6,
                    "Process Contingency": 0.12,
                    "Process Parameter": "Flue gas inlet to absorber",
                    "Project Contingency": 0.2,
                    "RP Value": 672504.35,
                    "Units": "m**3/hr",
                },
                "6.23.1.ccs": {
                    "Account Name": "MEA solvent capture system washing solvent tank cost component 1",
                    "BEC": 27.6,
                    "BEC_units": "K$2013",
                    "Eng Fee": 0.2,
                    "Exponent": 0.6,
                    "Process Contingency": 0.12,
                    "Process Parameter": "CO2 product flow rate",
                    "Project Contingency": 0.2,
                    "RP Value": 262349.78,
                    "Units": "lb/hr",
                },
                "6.23.2.ccs": {
                    "Account Name": "MEA solvent capture system washing solvent tank cost component 2",
                    "BEC": 18.4,
                    "BEC_units": "K$2013",
                    "Eng Fee": 0.2,
                    "Exponent": 0.6,
                    "Process Contingency": 0.12,
                    "Process Parameter": "Flue gas inlet to absorber",
                    "Project Contingency": 0.2,
                    "RP Value": 672504.35,
                    "Units": "m**3/hr",
                },
                "6.24.1.ccs": {
                    "Account Name": "MEA solvent capture system solvent stripper reclaimer cost component 1",
                    "BEC": 108.6,
                    "BEC_units": "K$2013",
                    "Eng Fee": 0.2,
                    "Exponent": 0.6,
                    "Process Contingency": 0.12,
                    "Process Parameter": "CO2 product flow rate",
                    "Project Contingency": 0.2,
                    "RP Value": 262349.78,
                    "Units": "lb/hr",
                },
                "6.24.2.ccs": {
                    "Account Name": "MEA solvent capture system solvent stripper reclaimer cost component 2",
                    "BEC": 72.4,
                    "BEC_units": "K$2013",
                    "Eng Fee": 0.2,
                    "Exponent": 0.6,
                    "Process Contingency": 0.12,
                    "Process Parameter": "Flue gas inlet to absorber",
                    "Project Contingency": 0.2,
                    "RP Value": 672504.35,
                    "Units": "m**3/hr",
                },
                "6.25.1.ccs": {
                    "Account Name": "MEA solvent capture system solvent reclaimer cooler cost component 1",
                    "BEC": 108.0,
                    "BEC_units": "K$2013",
                    "Eng Fee": 0.2,
                    "Exponent": 0.6,
                    "Process Contingency": 0.12,
                    "Process Parameter": "CO2 product flow rate",
                    "Project Contingency": 0.2,
                    "RP Value": 262349.78,
                    "Units": "lb/hr",
                },
                "6.25.2.ccs": {
                    "Account Name": "MEA solvent capture system solvent reclaimer cooler cost component 2",
                    "BEC": 72.0,
                    "BEC_units": "K$2013",
                    "Eng Fee": 0.2,
                    "Exponent": 0.6,
                    "Process Contingency": 0.12,
                    "Process Parameter": "Flue gas inlet to absorber",
                    "Project Contingency": 0.2,
                    "RP Value": 672504.35,
                    "Units": "m**3/hr",
                },
                "6.26.1.ccs": {
                    "Account Name": "MEA solvent capture system solvent filtration cost component 1",
                    "BEC": 519.6,
                    "BEC_units": "K$2013",
                    "Eng Fee": 0.2,
                    "Exponent": 0.6,
                    "Process Contingency": 0.12,
                    "Process Parameter": "CO2 product flow rate",
                    "Project Contingency": 0.2,
                    "RP Value": 262349.78,
                    "Units": "lb/hr",
                },
                "6.26.2.ccs": {
                    "Account Name": "MEA solvent capture system solvent filtration cost component 2",
                    "BEC": 346.4,
                    "BEC_units": "K$2013",
                    "Eng Fee": 0.2,
                    "Exponent": 0.6,
                    "Process Contingency": 0.12,
                    "Process Parameter": "Flue gas inlet to absorber",
                    "Project Contingency": 0.2,
                    "RP Value": 672504.35,
                    "Units": "m**3/hr",
                },
            }
        }
    }

    # check that file contains expected contents
    with open(os.path.join(wrkdir, "generic_ccs_costing_data.json"), "r") as file:
        json_contents = json.load(file)

    # check that function returns expected output
    assert json_contents == dictionaries_expected_output

    # should be the same as the check above
    assert generic_ccs_costing_params == dictionaries_expected_output

    _temp_context.release(remove=True)


@pytest.mark.skipif(solver is None, reason="Solver not available")
@pytest.mark.component
def test_ccs_units_costing():
    m = pyo.ConcreteModel()

    # Add a flowsheet object to the model
    m.fs = FlowsheetBlock(dynamic=False)
    m.fs.costing = QGESSCosting()

    # Accounts with carbon capture system units
    # Absorber
    absorber_accounts = ["6.1.ccs"]
    m.fs.b17 = pyo.Block()
    # Obtain absorber volume in m3
    absorber_volume = 4999  # m3

    m.fs.b17.absorber_volume = pyo.Var(initialize=absorber_volume, units=pyunits.m**3)
    m.fs.b17.absorber_volume.fix()
    QGESSCostingData.get_PP_costing(
        m.fs.b17,
        absorber_accounts,
        m.fs.b17.absorber_volume,
        6,
        CE_index_year="2013",
    )

    # Absorber packing
    absorber_packing_accounts = ["6.2.ccs"]
    m.fs.b18 = pyo.Block()
    # Obtain absorber packing volume in m3
    absorber_packing_volume = 3773  # m3

    m.fs.b18.absorber_packing_volume = pyo.Var(
        initialize=absorber_packing_volume, units=pyunits.m**3
    )
    m.fs.b18.absorber_packing_volume.fix()
    QGESSCostingData.get_PP_costing(
        m.fs.b18,
        absorber_packing_accounts,
        m.fs.b18.absorber_packing_volume,
        6,
        CE_index_year="2013",
    )

    # Stripper
    stripper_accounts = ["6.3.ccs"]
    m.fs.b19 = pyo.Block()
    # Obtain stripper volume in m3
    stripper_volume = 386  # m3

    m.fs.b19.stripper_volume = pyo.Var(initialize=stripper_volume, units=pyunits.m**3)
    m.fs.b19.stripper_volume.fix()
    QGESSCostingData.get_PP_costing(
        m.fs.b19,
        stripper_accounts,
        m.fs.b19.stripper_volume,
        6,
        CE_index_year="2013",
    )

    # Stripper packing
    stripper_packing_accounts = ["6.4.ccs"]
    m.fs.b20 = pyo.Block()
    # Obtain stripper packing volume in m3
    stripper_packing_volume = 291  # m3

    m.fs.b20.stripper_packing_volume = pyo.Var(
        initialize=stripper_packing_volume, units=pyunits.m**3
    )
    m.fs.b20.stripper_packing_volume.fix()
    QGESSCostingData.get_PP_costing(
        m.fs.b20,
        stripper_packing_accounts,
        m.fs.b20.stripper_packing_volume,
        6,
        CE_index_year="2013",
    )

    # Stripper condenser
    stripper_condenser_accounts = ["6.5.ccs"]
    m.fs.b21 = pyo.Block()
    # Obtain stripper condenser area in m2
    stripper_condenser_area = 1500  # m2

    m.fs.b21.stripper_condenser_area = pyo.Var(
        initialize=stripper_condenser_area, units=pyunits.m**2
    )
    m.fs.b21.stripper_condenser_area.fix()
    QGESSCostingData.get_PP_costing(
        m.fs.b21,
        stripper_condenser_accounts,
        m.fs.b21.stripper_condenser_area,
        6,
        CE_index_year="2013",
    )

    # Stripper reboiler
    stripper_reboiler_accounts = ["6.6.ccs"]
    m.fs.b22 = pyo.Block()
    # Obtain stripper reboiler area in m2
    stripper_reboiler_area = 1500  # m2

    m.fs.b22.stripper_reboiler_area = pyo.Var(
        initialize=stripper_reboiler_area, units=pyunits.m**2
    )
    m.fs.b22.stripper_reboiler_area.fix()
    QGESSCostingData.get_PP_costing(
        m.fs.b22,
        stripper_reboiler_accounts,
        m.fs.b22.stripper_reboiler_area,
        6,
        CE_index_year="2013",
    )

    # Lean rich heat exchanger
    lean_rich_hex_accounts = ["6.7.ccs"]
    m.fs.b23 = pyo.Block()
    # Obtain lean rich heat exchanger area in m2
    lean_rich_hex_area = 40700  # m2

    m.fs.b23.lean_rich_hex_area = pyo.Var(
        initialize=lean_rich_hex_area, units=pyunits.m**2
    )
    m.fs.b23.lean_rich_hex_area.fix()
    QGESSCostingData.get_PP_costing(
        m.fs.b23,
        lean_rich_hex_accounts,
        m.fs.b23.lean_rich_hex_area,
        6,
        CE_index_year="2013",
    )

    # Lean solvent cooler
    lean_solvent_cooler_accounts = ["6.8.ccs"]
    m.fs.b24 = pyo.Block()
    # Obtain lean solvent cooler area in m2
    lean_solvent_cooler_area = 600  # m2

    m.fs.b24.lean_solvent_cooler_area = pyo.Var(
        initialize=lean_solvent_cooler_area, units=pyunits.m**2
    )
    m.fs.b24.lean_solvent_cooler_area.fix()
    QGESSCostingData.get_PP_costing(
        m.fs.b24,
        lean_solvent_cooler_accounts,
        m.fs.b24.lean_solvent_cooler_area,
        6,
        CE_index_year="2013",
    )

    # Flue gas blower
    flue_gas_blower_account_1 = ["6.9.1.ccs"]
    flue_gas_blower_account_2 = ["6.9.2.ccs"]

    m.fs.b25 = pyo.Block()
    m.fs.b26 = pyo.Block()
    # Obtain the CO2 product flow in lb/hr
    co2_product_flow = 245000  # lb/hr
    fg_flow = 2e6  # m3/hr

    m.fs.b25.co2_product_flow = pyo.Var(
        initialize=co2_product_flow, units=pyunits.lb / pyunits.hr
    )
    m.fs.b25.co2_product_flow.fix()
    m.fs.b26.fg_flow = pyo.Var(initialize=fg_flow, units=pyunits.m**3 / pyunits.hr)
    m.fs.b26.fg_flow.fix()
    QGESSCostingData.get_PP_costing(
        m.fs.b25,
        flue_gas_blower_account_1,
        m.fs.b25.co2_product_flow,
        6,
        CE_index_year="2013",
    )
    QGESSCostingData.get_PP_costing(
        m.fs.b26,
        flue_gas_blower_account_2,
        m.fs.b26.fg_flow,
        6,
        CE_index_year="2013",
    )

    # Flue gas direct contact cooler
    flue_gas_dcc_account_1 = ["6.10.1.ccs"]
    flue_gas_dcc_account_2 = ["6.10.2.ccs"]
    m.fs.b27 = pyo.Block()
    m.fs.b28 = pyo.Block()

    m.fs.b27.co2_product_flow = pyo.Var(
        initialize=co2_product_flow, units=pyunits.lb / pyunits.hr
    )
    m.fs.b27.co2_product_flow.fix()
    m.fs.b28.fg_flow = pyo.Var(initialize=fg_flow, units=pyunits.m**3 / pyunits.hr)
    m.fs.b28.fg_flow.fix()
    QGESSCostingData.get_PP_costing(
        m.fs.b27,
        flue_gas_dcc_account_1,
        m.fs.b27.co2_product_flow,
        6,
        CE_index_year="2013",
    )
    QGESSCostingData.get_PP_costing(
        m.fs.b28,
        flue_gas_dcc_account_2,
        m.fs.b28.fg_flow,
        6,
        CE_index_year="2013",
    )

    # Flue gas direct contact cooler packing
    flue_gas_dcc_packing_account_1 = ["6.11.1.ccs"]
    flue_gas_dcc_packing_account_2 = ["6.11.2.ccs"]
    m.fs.b29 = pyo.Block()
    m.fs.b30 = pyo.Block()

    m.fs.b29.co2_product_flow = pyo.Var(
        initialize=co2_product_flow, units=pyunits.lb / pyunits.hr
    )
    m.fs.b29.co2_product_flow.fix()
    m.fs.b30.fg_flow = pyo.Var(initialize=fg_flow, units=pyunits.m**3 / pyunits.hr)
    m.fs.b30.fg_flow.fix()
    QGESSCostingData.get_PP_costing(
        m.fs.b29,
        flue_gas_dcc_packing_account_1,
        m.fs.b29.co2_product_flow,
        6,
        CE_index_year="2013",
    )
    QGESSCostingData.get_PP_costing(
        m.fs.b30,
        flue_gas_dcc_packing_account_2,
        m.fs.b30.fg_flow,
        6,
        CE_index_year="2013",
    )

    # Pretreatment pump
    pretreatment_pump_account_1 = ["6.12.1.ccs"]
    pretreatment_pump_account_2 = ["6.12.2.ccs"]
    m.fs.b31 = pyo.Block()
    m.fs.b32 = pyo.Block()

    m.fs.b31.co2_product_flow = pyo.Var(
        initialize=co2_product_flow, units=pyunits.lb / pyunits.hr
    )
    m.fs.b31.co2_product_flow.fix()
    m.fs.b32.fg_flow = pyo.Var(initialize=fg_flow, units=pyunits.m**3 / pyunits.hr)
    m.fs.b32.fg_flow.fix()
    QGESSCostingData.get_PP_costing(
        m.fs.b31,
        pretreatment_pump_account_1,
        m.fs.b31.co2_product_flow,
        6,
        CE_index_year="2013",
    )
    QGESSCostingData.get_PP_costing(
        m.fs.b32,
        pretreatment_pump_account_2,
        m.fs.b32.fg_flow,
        6,
        CE_index_year="2013",
    )

    # Pretreatment cooler
    pretreatment_cooler_account_1 = ["6.13.1.ccs"]
    pretreatment_cooler_account_2 = ["6.13.2.ccs"]
    m.fs.b33 = pyo.Block()
    m.fs.b34 = pyo.Block()

    m.fs.b33.co2_product_flow = pyo.Var(
        initialize=co2_product_flow, units=pyunits.lb / pyunits.hr
    )
    m.fs.b33.co2_product_flow.fix()
    m.fs.b34.fg_flow = pyo.Var(initialize=fg_flow, units=pyunits.m**3 / pyunits.hr)
    m.fs.b34.fg_flow.fix()
    QGESSCostingData.get_PP_costing(
        m.fs.b33,
        pretreatment_cooler_account_1,
        m.fs.b33.co2_product_flow,
        6,
        CE_index_year="2013",
    )
    QGESSCostingData.get_PP_costing(
        m.fs.b34,
        pretreatment_cooler_account_2,
        m.fs.b34.fg_flow,
        6,
        CE_index_year="2013",
    )

    # Pretreatment tank
    pretreatment_tank_account_1 = ["6.14.1.ccs"]
    pretreatment_tank_account_2 = ["6.14.2.ccs"]
    m.fs.b35 = pyo.Block()
    m.fs.b36 = pyo.Block()

    m.fs.b35.co2_product_flow = pyo.Var(
        initialize=co2_product_flow, units=pyunits.lb / pyunits.hr
    )
    m.fs.b35.co2_product_flow.fix()
    m.fs.b36.fg_flow = pyo.Var(initialize=fg_flow, units=pyunits.m**3 / pyunits.hr)
    m.fs.b36.fg_flow.fix()
    QGESSCostingData.get_PP_costing(
        m.fs.b35,
        pretreatment_tank_account_1,
        m.fs.b35.co2_product_flow,
        6,
        CE_index_year="2013",
    )
    QGESSCostingData.get_PP_costing(
        m.fs.b36,
        pretreatment_tank_account_2,
        m.fs.b36.fg_flow,
        6,
        CE_index_year="2013",
    )

    # Washing column
    washing_column_account_1 = ["6.15.1.ccs"]
    washing_column_account_2 = ["6.15.2.ccs"]
    m.fs.b37 = pyo.Block()
    m.fs.b38 = pyo.Block()

    m.fs.b37.co2_product_flow = pyo.Var(
        initialize=co2_product_flow, units=pyunits.lb / pyunits.hr
    )
    m.fs.b37.co2_product_flow.fix()
    m.fs.b38.fg_flow = pyo.Var(initialize=fg_flow, units=pyunits.m**3 / pyunits.hr)
    m.fs.b38.fg_flow.fix()
    QGESSCostingData.get_PP_costing(
        m.fs.b37,
        washing_column_account_1,
        m.fs.b37.co2_product_flow,
        6,
        CE_index_year="2013",
    )
    QGESSCostingData.get_PP_costing(
        m.fs.b38,
        washing_column_account_2,
        m.fs.b38.fg_flow,
        6,
        CE_index_year="2013",
    )

    # Washing column packing
    washing_column_packing_account_1 = ["6.16.1.ccs"]
    washing_column_packing_account_2 = ["6.16.2.ccs"]
    m.fs.b39 = pyo.Block()
    m.fs.b40 = pyo.Block()

    m.fs.b39.co2_product_flow = pyo.Var(
        initialize=co2_product_flow, units=pyunits.lb / pyunits.hr
    )
    m.fs.b39.co2_product_flow.fix()
    m.fs.b40.fg_flow = pyo.Var(initialize=fg_flow, units=pyunits.m**3 / pyunits.hr)
    m.fs.b40.fg_flow.fix()
    QGESSCostingData.get_PP_costing(
        m.fs.b39,
        washing_column_packing_account_1,
        m.fs.b39.co2_product_flow,
        6,
        CE_index_year="2013",
    )
    QGESSCostingData.get_PP_costing(
        m.fs.b40,
        washing_column_packing_account_2,
        m.fs.b40.fg_flow,
        6,
        CE_index_year="2013",
    )

    # Washing solvent cooler
    washing_solvent_cooler_account_1 = ["6.17.1.ccs"]
    washing_solvent_cooler_account_2 = ["6.17.2.ccs"]
    m.fs.b41 = pyo.Block()
    m.fs.b42 = pyo.Block()

    m.fs.b41.co2_product_flow = pyo.Var(
        initialize=co2_product_flow, units=pyunits.lb / pyunits.hr
    )
    m.fs.b41.co2_product_flow.fix()
    m.fs.b42.fg_flow = pyo.Var(initialize=fg_flow, units=pyunits.m**3 / pyunits.hr)
    m.fs.b42.fg_flow.fix()
    QGESSCostingData.get_PP_costing(
        m.fs.b41,
        washing_solvent_cooler_account_1,
        m.fs.b41.co2_product_flow,
        6,
        CE_index_year="2013",
    )
    QGESSCostingData.get_PP_costing(
        m.fs.b42,
        washing_solvent_cooler_account_2,
        m.fs.b42.fg_flow,
        6,
        CE_index_year="2013",
    )

    # Washing solvent pump
    washing_solvent_pump_account_1 = ["6.18.1.ccs"]
    washing_solvent_pump_account_2 = ["6.18.2.ccs"]
    m.fs.b43 = pyo.Block()
    m.fs.b44 = pyo.Block()

    m.fs.b43.co2_product_flow = pyo.Var(
        initialize=co2_product_flow, units=pyunits.lb / pyunits.hr
    )
    m.fs.b43.co2_product_flow.fix()
    m.fs.b44.fg_flow = pyo.Var(initialize=fg_flow, units=pyunits.m**3 / pyunits.hr)
    m.fs.b44.fg_flow.fix()
    QGESSCostingData.get_PP_costing(
        m.fs.b43,
        washing_solvent_pump_account_1,
        m.fs.b43.co2_product_flow,
        6,
        CE_index_year="2013",
    )
    QGESSCostingData.get_PP_costing(
        m.fs.b44,
        washing_solvent_pump_account_2,
        m.fs.b44.fg_flow,
        6,
        CE_index_year="2013",
    )

    # Condenser pump
    condenser_pump_account_1 = ["6.19.1.ccs"]
    condenser_pump_account_2 = ["6.19.2.ccs"]
    m.fs.b45 = pyo.Block()
    m.fs.b46 = pyo.Block()

    m.fs.b45.co2_product_flow = pyo.Var(
        initialize=co2_product_flow, units=pyunits.lb / pyunits.hr
    )
    m.fs.b45.co2_product_flow.fix()
    m.fs.b46.fg_flow = pyo.Var(initialize=fg_flow, units=pyunits.m**3 / pyunits.hr)
    m.fs.b46.fg_flow.fix()
    QGESSCostingData.get_PP_costing(
        m.fs.b45,
        condenser_pump_account_1,
        m.fs.b45.co2_product_flow,
        6,
        CE_index_year="2013",
    )
    QGESSCostingData.get_PP_costing(
        m.fs.b46,
        condenser_pump_account_2,
        m.fs.b46.fg_flow,
        6,
        CE_index_year="2013",
    )

    # Stripper reflux drum
    stripper_reflux_drum_account_1 = ["6.20.1.ccs"]
    stripper_reflux_drum_account_2 = ["6.20.2.ccs"]
    m.fs.b47 = pyo.Block()
    m.fs.b48 = pyo.Block()

    m.fs.b47.co2_product_flow = pyo.Var(
        initialize=co2_product_flow, units=pyunits.lb / pyunits.hr
    )
    m.fs.b47.co2_product_flow.fix()
    m.fs.b48.fg_flow = pyo.Var(initialize=fg_flow, units=pyunits.m**3 / pyunits.hr)
    m.fs.b48.fg_flow.fix()
    QGESSCostingData.get_PP_costing(
        m.fs.b47,
        stripper_reflux_drum_account_1,
        m.fs.b47.co2_product_flow,
        6,
        CE_index_year="2013",
    )
    QGESSCostingData.get_PP_costing(
        m.fs.b48,
        stripper_reflux_drum_account_2,
        m.fs.b48.fg_flow,
        6,
        CE_index_year="2013",
    )

    # Lean solvent pump
    lean_solvent_pump_account_1 = ["6.21.1.ccs"]
    lean_solvent_pump_account_2 = ["6.21.2.ccs"]
    m.fs.b49 = pyo.Block()
    m.fs.b50 = pyo.Block()

    m.fs.b49.co2_product_flow = pyo.Var(
        initialize=co2_product_flow, units=pyunits.lb / pyunits.hr
    )
    m.fs.b49.co2_product_flow.fix()
    m.fs.b50.fg_flow = pyo.Var(initialize=fg_flow, units=pyunits.m**3 / pyunits.hr)
    m.fs.b50.fg_flow.fix()
    QGESSCostingData.get_PP_costing(
        m.fs.b49,
        lean_solvent_pump_account_1,
        m.fs.b49.co2_product_flow,
        6,
        CE_index_year="2013",
    )
    QGESSCostingData.get_PP_costing(
        m.fs.b50,
        lean_solvent_pump_account_2,
        m.fs.b50.fg_flow,
        6,
        CE_index_year="2013",
    )

    # Solvent storage tank
    solvent_storage_tank_account_1 = ["6.22.1.ccs"]
    solvent_storage_tank_account_2 = ["6.22.2.ccs"]
    m.fs.b51 = pyo.Block()
    m.fs.b52 = pyo.Block()

    m.fs.b51.co2_product_flow = pyo.Var(
        initialize=co2_product_flow, units=pyunits.lb / pyunits.hr
    )
    m.fs.b51.co2_product_flow.fix()
    m.fs.b52.fg_flow = pyo.Var(initialize=fg_flow, units=pyunits.m**3 / pyunits.hr)
    m.fs.b52.fg_flow.fix()
    QGESSCostingData.get_PP_costing(
        m.fs.b51,
        solvent_storage_tank_account_1,
        m.fs.b51.co2_product_flow,
        6,
        CE_index_year="2013",
    )
    QGESSCostingData.get_PP_costing(
        m.fs.b52,
        solvent_storage_tank_account_2,
        m.fs.b52.fg_flow,
        6,
        CE_index_year="2013",
    )

    # Washing solvent tank
    washing_solvent_tank_account_1 = ["6.23.1.ccs"]
    washing_solvent_tank_account_2 = ["6.23.2.ccs"]
    m.fs.b53 = pyo.Block()
    m.fs.b54 = pyo.Block()

    m.fs.b53.co2_product_flow = pyo.Var(
        initialize=co2_product_flow, units=pyunits.lb / pyunits.hr
    )
    m.fs.b53.co2_product_flow.fix()
    m.fs.b54.fg_flow = pyo.Var(initialize=fg_flow, units=pyunits.m**3 / pyunits.hr)
    m.fs.b54.fg_flow.fix()
    QGESSCostingData.get_PP_costing(
        m.fs.b53,
        washing_solvent_tank_account_1,
        m.fs.b53.co2_product_flow,
        6,
        CE_index_year="2013",
    )
    QGESSCostingData.get_PP_costing(
        m.fs.b54,
        washing_solvent_tank_account_2,
        m.fs.b54.fg_flow,
        6,
        CE_index_year="2013",
    )

    # Solvent stripper reclaimer
    solvent_stripper_reclaimer_account_1 = ["6.24.1.ccs"]
    solvent_stripper_reclaimer_account_2 = ["6.24.2.ccs"]
    m.fs.b55 = pyo.Block()
    m.fs.b56 = pyo.Block()

    m.fs.b55.co2_product_flow = pyo.Var(
        initialize=co2_product_flow, units=pyunits.lb / pyunits.hr
    )
    m.fs.b55.co2_product_flow.fix()
    m.fs.b56.fg_flow = pyo.Var(initialize=fg_flow, units=pyunits.m**3 / pyunits.hr)
    m.fs.b56.fg_flow.fix()
    QGESSCostingData.get_PP_costing(
        m.fs.b55,
        solvent_stripper_reclaimer_account_1,
        m.fs.b55.co2_product_flow,
        6,
        CE_index_year="2013",
    )
    QGESSCostingData.get_PP_costing(
        m.fs.b56,
        solvent_stripper_reclaimer_account_2,
        m.fs.b56.fg_flow,
        6,
        CE_index_year="2013",
    )

    # Solvent reclaimer cooler
    solvent_reclaimer_cooler_account_1 = ["6.25.1.ccs"]
    solvent_reclaimer_cooler_account_2 = ["6.25.2.ccs"]
    m.fs.b57 = pyo.Block()
    m.fs.b58 = pyo.Block()

    m.fs.b57.co2_product_flow = pyo.Var(
        initialize=co2_product_flow, units=pyunits.lb / pyunits.hr
    )
    m.fs.b57.co2_product_flow.fix()
    m.fs.b58.fg_flow = pyo.Var(initialize=fg_flow, units=pyunits.m**3 / pyunits.hr)
    m.fs.b58.fg_flow.fix()
    QGESSCostingData.get_PP_costing(
        m.fs.b57,
        solvent_reclaimer_cooler_account_1,
        m.fs.b57.co2_product_flow,
        6,
        CE_index_year="2013",
    )
    QGESSCostingData.get_PP_costing(
        m.fs.b58,
        solvent_reclaimer_cooler_account_2,
        m.fs.b58.fg_flow,
        6,
        CE_index_year="2013",
    )

    # Solvent filtration
    solvent_filtration_account_1 = ["6.26.1.ccs"]
    solvent_filtration_account_2 = ["6.26.2.ccs"]
    m.fs.b59 = pyo.Block()
    m.fs.b60 = pyo.Block()

    m.fs.b59.co2_product_flow = pyo.Var(
        initialize=co2_product_flow, units=pyunits.lb / pyunits.hr
    )
    m.fs.b59.co2_product_flow.fix()
    m.fs.b60.fg_flow = pyo.Var(initialize=fg_flow, units=pyunits.m**3 / pyunits.hr)
    m.fs.b60.fg_flow.fix()
    QGESSCostingData.get_PP_costing(
        m.fs.b59,
        solvent_filtration_account_1,
        m.fs.b59.co2_product_flow,
        6,
        CE_index_year="2013",
    )
    QGESSCostingData.get_PP_costing(
        m.fs.b60,
        solvent_filtration_account_2,
        m.fs.b60.fg_flow,
        6,
        CE_index_year="2013",
    )

    m.fs.costing.build_process_costs(
        net_power=None,
        fixed_OM=False,
        variable_OM=False,
        resources=None,
        rates=None,
        prices=None,
        fuel=None,
        CE_index_year="2013",
    )

    # Initialize costing
    QGESSCostingData.costing_initialization(m.fs.costing)
    assert degrees_of_freedom(m) == 0

    # Solve the model
    results = solver.solve(m, tee=True)

    # check unit consistency
    assert_units_consistent(m)

    assert results.solver.termination_condition == pyo.TerminationCondition.optimal
    assert results.solver.status == pyo.SolverStatus.ok

    # Absorber TPC
    assert pytest.approx(33.375, rel=1e-3) == sum(
        pyo.value(m.fs.b17.total_plant_cost[ac]) for ac in absorber_accounts
    )

    # Absorber packing TPC
    assert pytest.approx(24.727, rel=1e-3) == sum(
        pyo.value(m.fs.b18.total_plant_cost[ac]) for ac in absorber_packing_accounts
    )

    # Stripper TPC
    assert pytest.approx(3.424, rel=1e-3) == sum(
        pyo.value(m.fs.b19.total_plant_cost[ac]) for ac in stripper_accounts
    )

    # Stripper packing TPC
    assert pytest.approx(2.236, rel=1e-3) == sum(
        pyo.value(m.fs.b20.total_plant_cost[ac]) for ac in stripper_packing_accounts
    )

    # Stripper condenser TPC
    assert pytest.approx(0.7045, rel=1e-3) == sum(
        pyo.value(m.fs.b21.total_plant_cost[ac]) for ac in stripper_condenser_accounts
    )

    # Stripper reboiler TPC
    assert pytest.approx(2.699, rel=1e-3) == sum(
        pyo.value(m.fs.b22.total_plant_cost[ac]) for ac in stripper_reboiler_accounts
    )

    # Lean rich heat exchanger TPC
    assert pytest.approx(5.845, rel=1e-3) == sum(
        pyo.value(m.fs.b23.total_plant_cost[ac]) for ac in lean_rich_hex_accounts
    )

    # Lean solvent cooler TPC
    assert pytest.approx(0.5455, rel=1e-3) == sum(
        pyo.value(m.fs.b24.total_plant_cost[ac]) for ac in lean_solvent_cooler_accounts
    )

    # Flue gas blower TPC
    assert pytest.approx(0.7407, rel=1e-3) == sum(
        pyo.value(m.fs.b25.total_plant_cost[ac]) for ac in flue_gas_blower_account_1
    )
    assert pytest.approx(0.9894, rel=1e-3) == sum(
        pyo.value(m.fs.b26.total_plant_cost[ac]) for ac in flue_gas_blower_account_2
    )

    # Flue gas direct contact cooler TPC
    assert pytest.approx(2.623, rel=1e-3) == sum(
        pyo.value(m.fs.b27.total_plant_cost[ac]) for ac in flue_gas_dcc_account_1
    )
    assert pytest.approx(3.503, rel=1e-3) == sum(
        pyo.value(m.fs.b28.total_plant_cost[ac]) for ac in flue_gas_dcc_account_2
    )

    # Flue gas direct contact cooler packing TPC
    assert pytest.approx(1.718, rel=1e-3) == sum(
        pyo.value(m.fs.b29.total_plant_cost[ac])
        for ac in flue_gas_dcc_packing_account_1
    )
    assert pytest.approx(2.294, rel=1e-3) == sum(
        pyo.value(m.fs.b30.total_plant_cost[ac])
        for ac in flue_gas_dcc_packing_account_2
    )

    # Pretreatment pump TPC
    assert pytest.approx(0.09851, rel=1e-3) == sum(
        pyo.value(m.fs.b31.total_plant_cost[ac]) for ac in pretreatment_pump_account_1
    )
    assert pytest.approx(0.1316, rel=1e-3) == sum(
        pyo.value(m.fs.b32.total_plant_cost[ac]) for ac in pretreatment_pump_account_2
    )

    # Pretreatment cooler TPC
    assert pytest.approx(0.1843, rel=1e-3) == sum(
        pyo.value(m.fs.b33.total_plant_cost[ac]) for ac in pretreatment_cooler_account_1
    )
    assert pytest.approx(0.2461, rel=1e-3) == sum(
        pyo.value(m.fs.b34.total_plant_cost[ac]) for ac in pretreatment_cooler_account_2
    )

    # Pretreatment tank TPC
    assert pytest.approx(0.08483, rel=1e-3) == sum(
        pyo.value(m.fs.b35.total_plant_cost[ac]) for ac in pretreatment_tank_account_1
    )
    assert pytest.approx(0.1133, rel=1e-3) == sum(
        pyo.value(m.fs.b36.total_plant_cost[ac]) for ac in pretreatment_tank_account_2
    )

    # Washing column TPC
    assert pytest.approx(2.594, rel=1e-3) == sum(
        pyo.value(m.fs.b37.total_plant_cost[ac]) for ac in washing_column_account_1
    )
    assert pytest.approx(3.463, rel=1e-3) == sum(
        pyo.value(m.fs.b38.total_plant_cost[ac]) for ac in washing_column_account_2
    )

    # Washing column packing TPC
    assert pytest.approx(1.879, rel=1e-3) == sum(
        pyo.value(m.fs.b39.total_plant_cost[ac])
        for ac in washing_column_packing_account_1
    )
    assert pytest.approx(2.510, rel=1e-3) == sum(
        pyo.value(m.fs.b40.total_plant_cost[ac])
        for ac in washing_column_packing_account_2
    )

    # Washing solvent cooler TPC
    assert pytest.approx(0.06294, rel=1e-3) == sum(
        pyo.value(m.fs.b41.total_plant_cost[ac])
        for ac in washing_solvent_cooler_account_1
    )
    assert pytest.approx(0.08407, rel=1e-3) == sum(
        pyo.value(m.fs.b42.total_plant_cost[ac])
        for ac in washing_solvent_cooler_account_2
    )

    # Washing solvent pump TPC
    assert pytest.approx(0.01551, rel=1e-3) == sum(
        pyo.value(m.fs.b43.total_plant_cost[ac])
        for ac in washing_solvent_pump_account_1
    )
    assert pytest.approx(0.02071, rel=1e-3) == sum(
        pyo.value(m.fs.b44.total_plant_cost[ac])
        for ac in washing_solvent_pump_account_2
    )

    # Condenser pump TPC
    assert pytest.approx(0.03193, rel=1e-3) == sum(
        pyo.value(m.fs.b45.total_plant_cost[ac]) for ac in condenser_pump_account_1
    )
    assert pytest.approx(0.04265, rel=1e-3) == sum(
        pyo.value(m.fs.b46.total_plant_cost[ac]) for ac in condenser_pump_account_2
    )

    # Stripper reflux drum TPC
    assert pytest.approx(0.04196, rel=1e-3) == sum(
        pyo.value(m.fs.b47.total_plant_cost[ac])
        for ac in stripper_reflux_drum_account_1
    )
    assert pytest.approx(0.05605, rel=1e-3) == sum(
        pyo.value(m.fs.b48.total_plant_cost[ac])
        for ac in stripper_reflux_drum_account_2
    )

    # Lean solvent pump TPC
    assert pytest.approx(0.1450, rel=1e-3) == sum(
        pyo.value(m.fs.b49.total_plant_cost[ac]) for ac in lean_solvent_pump_account_1
    )
    assert pytest.approx(0.1937, rel=1e-3) == sum(
        pyo.value(m.fs.b50.total_plant_cost[ac]) for ac in lean_solvent_pump_account_2
    )

    # Solvent storage tank TPC
    assert pytest.approx(0.3120, rel=1e-3) == sum(
        pyo.value(m.fs.b51.total_plant_cost[ac])
        for ac in solvent_storage_tank_account_1
    )
    assert pytest.approx(0.4167, rel=1e-3) == sum(
        pyo.value(m.fs.b52.total_plant_cost[ac])
        for ac in solvent_storage_tank_account_2
    )

    # Washing solvent tank TPC
    assert pytest.approx(0.04196, rel=1e-3) == sum(
        pyo.value(m.fs.b53.total_plant_cost[ac])
        for ac in washing_solvent_tank_account_1
    )
    assert pytest.approx(0.05605, rel=1e-3) == sum(
        pyo.value(m.fs.b54.total_plant_cost[ac])
        for ac in washing_solvent_tank_account_2
    )

    # Solvent stripper reclaimer TPC
    assert pytest.approx(0.1651, rel=1e-3) == sum(
        pyo.value(m.fs.b55.total_plant_cost[ac])
        for ac in solvent_stripper_reclaimer_account_1
    )
    assert pytest.approx(0.2205, rel=1e-3) == sum(
        pyo.value(m.fs.b56.total_plant_cost[ac])
        for ac in solvent_stripper_reclaimer_account_2
    )

    # Solvent reclaimer cooler TPC
    assert pytest.approx(0.1642, rel=1e-3) == sum(
        pyo.value(m.fs.b57.total_plant_cost[ac])
        for ac in solvent_reclaimer_cooler_account_1
    )
    assert pytest.approx(0.2193, rel=1e-3) == sum(
        pyo.value(m.fs.b58.total_plant_cost[ac])
        for ac in solvent_reclaimer_cooler_account_2
    )

    # Solvent filtration TPC
    assert pytest.approx(0.7899, rel=1e-3) == sum(
        pyo.value(m.fs.b59.total_plant_cost[ac]) for ac in solvent_filtration_account_1
    )
    assert pytest.approx(1.055, rel=1e-3) == sum(
        pyo.value(m.fs.b60.total_plant_cost[ac]) for ac in solvent_filtration_account_2
    )

    # Total TPC
    TPC = 0
    for idx in range(17, 61):
        block = getattr(m.fs, "b" + str(idx))
        for ac in block.total_plant_cost.keys():
            TPC += pyo.value(block.total_plant_cost[ac])

    assert pytest.approx(100.86, rel=1e-3) == TPC



#################################################################################
# The Institute for the Design of Advanced Energy Systems Integrated Platform
# Framework (IDAES IP) was produced under the DOE Institute for the
# Design of Advanced Energy Systems (IDAES).
#
# Copyright (c) 2018-2024 by the software owners: The Regents of the
# University of California, through Lawrence Berkeley National Laboratory,
# National Technology & Engineering Solutions of Sandia, LLC, Carnegie Mellon
# University, West Virginia University Research Corporation, et al.
# All rights reserved.  Please see the files COPYRIGHT.md and LICENSE.md
# for full copyright and license information.
#################################################################################
"""
Tests for generic property package core code

Author: Andrew Lee
"""
import functools
import pytest
import re
from sys import modules
from types import MethodType

from pyomo.environ import value, Block, ConcreteModel, Param, Set, Var, units as pyunits
from pyomo.util.check_units import assert_units_equivalent

from idaes.models.properties.modular_properties.base.generic_property import (
    GenericParameterBlock,
    GenericParameterData,
    GenericStateBlock,
    _initialize_critical_props,
)
from idaes.models.properties.modular_properties.base.tests.dummy_eos import DummyEoS

from idaes.core import (
    declare_process_block_class,
    Component,
    FlowsheetBlock,
    Phase,
    LiquidPhase,
    VaporPhase,
    MaterialBalanceType,
    MaterialFlowBasis,
    Solvent,
    PhaseType as PT,
)
from idaes.core.util.exceptions import ConfigurationError, PropertyPackageError
from idaes.models.properties.modular_properties.phase_equil.henry import HenryType
from idaes.models.properties.modular_properties.eos.ceos import Cubic, CubicType
from idaes.models.properties.modular_properties.state_definitions import FTPx
from idaes.core.base.property_meta import UnitSet
from idaes.core.initialization import BlockTriangularizationInitializer

from idaes.models.properties.modular_properties.phase_equil.henry import HenryType
from idaes.models.properties.modular_properties.examples.BT_ideal import (
    configuration as BTconfig,
)
from idaes.models.unit_models.flash import Flash

import idaes.logger as idaeslog


# -----------------------------------------------------------------------------
# Dummy method to avoid errors when setting metadata dict
def set_metadata(b):
    pass


def calculate_scaling_factors(b):
    b.scaling_check = True


# Dummy build_parameter methods for tests
def build_parameters(cobj, p, *args, **kwargs):
    cobj.add_component("test_param_" + p, Var(initialize=42))


# Declare a base units dict to save code later
base_units = {
    "time": pyunits.s,
    "length": pyunits.m,
    "mass": pyunits.kg,
    "amount": pyunits.mol,
    "temperature": pyunits.K,
}


@declare_process_block_class("DummyParameterBlock")
class DummyParameterData(GenericParameterData):
    def configure(self):
        self.configured = True


class TestGenericParameterBlock(object):
    @pytest.mark.unit
    def test_build(self):
        m = ConcreteModel()
        m.params = DummyParameterBlock(
            components={"a": {}, "b": {}, "c": {}},
            phases={
                "p1": {
                    "type": LiquidPhase,
                    "component_list": ["a", "b"],
                    "equation_of_state": DummyEoS,
                },
                "p2": {"equation_of_state": DummyEoS},
            },
            state_definition=modules[__name__],
            pressure_ref=100000.0,
            temperature_ref=300,
            base_units=base_units,
        )

        assert m.params.configured

        default_units = m.params.get_metadata().default_units
        assert isinstance(default_units, UnitSet)
        assert_units_equivalent(default_units.TIME, pyunits.s)
        assert_units_equivalent(default_units.LENGTH, pyunits.m)
        assert_units_equivalent(default_units.MASS, pyunits.kg)
        assert_units_equivalent(default_units.AMOUNT, pyunits.mol)
        assert_units_equivalent(default_units.TEMPERATURE, pyunits.K)
        assert_units_equivalent(default_units.CURRENT, pyunits.ampere)
        assert_units_equivalent(default_units.LUMINOUS_INTENSITY, pyunits.candela)

        assert isinstance(m.params.component_list, Set)
        assert len(m.params.component_list) == 3
        for j in m.params.component_list:
            assert j in ["a", "b", "c"]
            assert isinstance(m.params.get_component(j), Component)

        assert isinstance(m.params.phase_list, Set)
        assert len(m.params.phase_list) == 2
        for p in m.params.phase_list:
            assert p in ["p1", "p2"]
        assert isinstance(m.params.get_phase("p1"), LiquidPhase)
        assert isinstance(m.params.get_phase("p2"), Phase)
        assert m.params.p1.config.equation_of_state == DummyEoS
        assert m.params.p2.config.equation_of_state == DummyEoS

        assert isinstance(m.params._phase_component_set, Set)
        assert len(m.params._phase_component_set) == 5
        for i in m.params._phase_component_set:
            assert i in [
                ("p1", "a"),
                ("p1", "b"),
                ("p2", "a"),
                ("p2", "b"),
                ("p2", "c"),
            ]

        assert isinstance(m.params.pressure_ref, Param)
        assert m.params.pressure_ref.value == 1e5
        assert isinstance(m.params.temperature_ref, Param)
        assert m.params.temperature_ref.value == 300

        assert m.params.state_block_class is GenericStateBlock

        assert len(m.params.config.inherent_reactions) == 0
        assert m.params.config.reaction_basis == MaterialFlowBasis.molar

        assert not m.params.has_inherent_reactions

    @pytest.mark.unit
    def test_invalid_unit(self):
        m = ConcreteModel()

        with pytest.raises(
            PropertyPackageError,
            match=re.escape(
                "Unrecognized units of measurement for quantity TIME (foo)"
            ),
        ):
            m.params = DummyParameterBlock(
                components={"a": {}, "b": {}, "c": {}},
                phases={
                    "p1": {
                        "type": LiquidPhase,
                        "component_list": ["a", "b"],
                        "equation_of_state": DummyEoS,
                    },
                    "p2": {"equation_of_state": DummyEoS},
                },
                state_definition=modules[__name__],
                pressure_ref=100000.0,
                temperature_ref=300,
                base_units={
                    "time": "foo",
                    "length": pyunits.m,
                    "mass": pyunits.kg,
                    "amount": pyunits.mol,
                    "temperature": pyunits.K,
                },
            )

    @pytest.mark.unit
    def test_no_components(self):
        m = ConcreteModel()

        with pytest.raises(
            ConfigurationError,
            match="params was not provided with a components argument.",
        ):
            m.params = DummyParameterBlock(
                phases={
                    "p1": {"equation_of_state": "foo"},
                    "p2": {"equation_of_state": "bar"},
                },
                base_units=base_units,
            )

    @pytest.mark.unit
    def test_no_phases(self):
        m = ConcreteModel()

        with pytest.raises(
            ConfigurationError,
            match="params was not provided with a phases argument. "
            "Did you forget to unpack the configurations dictionary?",
        ):
            m.params = DummyParameterBlock(
                components={"a": {}, "b": {}, "c": {}}, base_units=base_units
            )

    @pytest.mark.unit
    def test_packed_dict(self):
        m = ConcreteModel()

        dummy_dict = {
            "phases": {
                "p1": {"equation_of_state": "foo"},
                "p2": {"equation_of_state": "bar"},
            },
        }

        with pytest.raises(
            ConfigurationError,
            match=re.escape(
                "params[phases] was not provided with a phases argument. "
                "Did you forget to unpack the configurations dictionary?"
            ),
        ):
            m.params = DummyParameterBlock(dummy_dict)

    @pytest.mark.unit
    def test_invalid_component_in_phase_component_list(self):
        m = ConcreteModel()

        with pytest.raises(
            ConfigurationError,
            match="params phase-component list for phase p1 "
            "contained component d which is not in the master "
            "component list",
        ):
            m.params = DummyParameterBlock(
                components={"a": {}, "b": {}, "c": {}},
                phases={
                    "p1": {
                        "type": LiquidPhase,
                        "component_list": ["a", "d"],
                        "equation_of_state": "foo",
                    },
                    "p2": {"equation_of_state": "bar"},
                },
                base_units=base_units,
            )

    @pytest.mark.unit
    def test_invalid_orphaned_component_in_phase_component_list(self):
        m = ConcreteModel()

        with pytest.raises(
            ConfigurationError,
            match="params Component c does not appear to be "
            "valid in any phase. Please check the component "
            "lists defined for each phase, and be sure you do "
            "not have generic Components in single-phase "
            "aqueous systems.",
        ):
            m.params = DummyParameterBlock(
                components={"a": {}, "b": {}, "c": {}},
                phases={
                    "p1": {
                        "type": LiquidPhase,
                        "component_list": ["a", "b"],
                        "equation_of_state": "foo",
                    }
                },
                base_units=base_units,
            )

    @pytest.mark.unit
    def test_invalid_component_in_phase_component_list_2(self):
        m = ConcreteModel()

        with pytest.raises(
            ConfigurationError,
            match="params phase-component list for phase p1 "
            "contained component a, however this component is "
            "not valid for the given PhaseType",
        ):
            m.params = DummyParameterBlock(
                components={
                    "a": {"valid_phase_types": PT.solidPhase},
                    "b": {},
                    "c": {},
                },
                phases={
                    "p1": {
                        "type": LiquidPhase,
                        "component_list": ["a", "b"],
                        "equation_of_state": "foo",
                    },
                    "p2": {"equation_of_state": "bar"},
                },
                base_units=base_units,
            )

    @pytest.mark.unit
    def test_phase_component_set_from_valid_phases(self):
        m = ConcreteModel()

        m.params = DummyParameterBlock(
            components={
                "a": {"valid_phase_types": PT.liquidPhase},
                "b": {"valid_phase_types": PT.vaporPhase},
                "c": {"valid_phase_types": [PT.liquidPhase, PT.vaporPhase]},
            },
            phases={
                "p1": {"type": LiquidPhase, "equation_of_state": DummyEoS},
                "p2": {"type": VaporPhase, "equation_of_state": DummyEoS},
            },
            state_definition=modules[__name__],
            pressure_ref=100000.0,
            temperature_ref=300,
            base_units=base_units,
        )

        assert len(m.params._phase_component_set) == 4
        for i in m.params._phase_component_set:
            assert i in [("p1", "a"), ("p1", "c"), ("p2", "b"), ("p2", "c")]

    @pytest.mark.unit
    def test_no_state_definition(self):
        m = ConcreteModel()

        with pytest.raises(
            ConfigurationError,
            match="params Generic Property Package was not "
            "provided with a state_definition configuration "
            "argument. Please fix your property parameter "
            "definition to include this.",
        ):
            m.params = DummyParameterBlock(
                components={"a": {}, "b": {}, "c": {}},
                phases={
                    "p1": {"equation_of_state": "foo"},
                    "p2": {"equation_of_state": "bar"},
                },
                base_units=base_units,
            )

    @pytest.mark.unit
    def test_no_pressure_ref(self):
        m = ConcreteModel()

        with pytest.raises(
            ConfigurationError,
            match="params Generic Property Package was not "
            "provided with a pressure_ref configuration "
            "argument. Please fix your property parameter "
            "definition to include this.",
        ):
            m.params = DummyParameterBlock(
                components={"a": {}, "b": {}, "c": {}},
                phases={"p1": {"equation_of_state": "foo"}, "p2": {}},
                state_definition="baz",
                base_units=base_units,
            )

    @pytest.mark.unit
    def test_convert_pressure_ref(self):
        m = ConcreteModel()

        # This will fail, but should set the reference pressure
        with pytest.raises(ConfigurationError):
            m.params = DummyParameterBlock(
                components={"a": {}, "b": {}, "c": {}},
                phases={"p1": {"equation_of_state": "foo"}, "p2": {}},
                state_definition="baz",
                pressure_ref=(1, pyunits.bar),
                base_units=base_units,
            )
        assert m.params.pressure_ref.value == 1e5

    @pytest.mark.unit
    def test_log_no_units_pressure_ref(self, caplog):
        m = ConcreteModel()

        caplog.set_level(
            idaeslog.DEBUG,
            logger=("idaes.models.properties.modular_properties.base.generic_property"),
        )

        # This will fail, but should set the reference pressure
        with pytest.raises(ConfigurationError):
            m.params = DummyParameterBlock(
                components={"a": {}, "b": {}, "c": {}},
                phases={"p1": {"equation_of_state": "foo"}, "p2": {}},
                state_definition="baz",
                pressure_ref=100000.0,
                base_units=base_units,
            )
        assert m.params.pressure_ref.value == 1e5
        assert (
            "params no units provided for parameter pressure_ref - "
            "assuming default units" in caplog.text
        )

    @pytest.mark.unit
    def test_no_temperature_ref(self):
        m = ConcreteModel()

        with pytest.raises(
            ConfigurationError,
            match="params Generic Property Package was not "
            "provided with a temperature_ref configuration "
            "argument. Please fix your property parameter "
            "definition to include this.",
        ):
            m.params = DummyParameterBlock(
                components={"a": {}, "b": {}, "c": {}},
                phases={"p1": {"equation_of_state": "foo"}, "p2": {}},
                state_definition="baz",
                pressure_ref=100000.0,
                base_units=base_units,
            )

    @pytest.mark.unit
    def test_log_no_units_temperature_ref(self, caplog):
        m = ConcreteModel()

        caplog.set_level(
            idaeslog.DEBUG,
            logger=("idaes.models.properties.modular_properties.base.generic_property"),
        )

        # This will fail, but should set the reference pressure
        with pytest.raises(ConfigurationError):
            m.params = DummyParameterBlock(
                components={"a": {}, "b": {}, "c": {}},
                phases={"p1": {"equation_of_state": "foo"}, "p2": {}},
                state_definition="baz",
                pressure_ref=100000.0,
                temperature_ref=300,
                base_units=base_units,
            )
        assert m.params.temperature_ref.value == 300
        assert (
            "params no units provided for parameter pressure_ref - "
            "assuming default units" in caplog.text
        )

    @pytest.mark.unit
    def test_convert_temperature_ref(self):
        m = ConcreteModel()

        # This will fail, but should set temerpature_ref
        with pytest.raises(ConfigurationError):
            m.params = DummyParameterBlock(
                components={"a": {}, "b": {}, "c": {}},
                phases={"p1": {"equation_of_state": "foo"}, "p2": {}},
                state_definition="baz",
                pressure_ref=100000.0,
                temperature_ref=(540, pyunits.degR),
                base_units=base_units,
            )
        assert m.params.temperature_ref.value == 300

    @pytest.mark.unit
    def test_no_eos(self):
        m = ConcreteModel()

        with pytest.raises(
            ConfigurationError,
            match="params phase p2 was not provided with an "
            "equation_of_state configuration argument. Please "
            "fix your property parameter definition to "
            "include this.",
        ):
            m.params = DummyParameterBlock(
                components={"a": {}, "b": {}, "c": {}},
                phases={"p1": {"equation_of_state": "foo"}, "p2": {}},
                state_definition="baz",
                pressure_ref=100000.0,
                temperature_ref=300,
                base_units=base_units,
            )

    @pytest.mark.unit
    def test_phases_in_equilibrium(self):
        m = ConcreteModel()

        m.params = DummyParameterBlock(
            components={
                "a": {"phase_equilibrium_form": {("p1", "p2"): "foo"}},
                "b": {"phase_equilibrium_form": {("p1", "p2"): "foo"}},
                "c": {"phase_equilibrium_form": {("p1", "p2"): "foo"}},
            },
            phases={
                "p1": {"equation_of_state": DummyEoS},
                "p2": {"equation_of_state": DummyEoS},
            },
            state_definition=modules[__name__],
            pressure_ref=100000.0,
            temperature_ref=300,
            phases_in_equilibrium=[("p1", "p2")],
            phase_equilibrium_state={("p1", "p2"): "whoop"},
            base_units=base_units,
        )

        assert isinstance(m.params.phase_equilibrium_idx, Set)
        assert len(m.params.phase_equilibrium_idx) == 3
        for i in m.params.phase_equilibrium_idx:
            assert i in ["PE1", "PE2", "PE3"]

        assert isinstance(m.params.phase_equilibrium_list, dict)
        assert m.params.phase_equilibrium_list == {
            "PE1": ["a", ("p1", "p2")],
            "PE2": ["b", ("p1", "p2")],
            "PE3": ["c", ("p1", "p2")],
        }

    @pytest.mark.unit
    def test_phases_in_equilibrium_no_form(self):
        m = ConcreteModel()

        with pytest.raises(
            ConfigurationError,
            match="params Generic Property Package component a "
            "is in equilibrium but phase_equilibrium_form"
            " was not specified.",
        ):
            m.params = DummyParameterBlock(
                components={"a": {}, "b": {}, "c": {}},
                phases={
                    "p1": {"equation_of_state": "foo"},
                    "p2": {"equation_of_state": "bar"},
                },
                state_definition="baz",
                pressure_ref=100000.0,
                temperature_ref=300,
                phases_in_equilibrium=[("p1", "p2")],
                phase_equilibrium_state={("p1", "p2"): "whoop"},
                base_units=base_units,
            )

    @pytest.mark.unit
    def test_phases_in_equilibrium_missing_pair_form(self):
        m = ConcreteModel()

        with pytest.raises(
            ConfigurationError,
            match="params Generic Property Package component b "
            "is in equilibrium but phase_equilibrium_form"
            " was not specified for all appropriate "
            "phase pairs.",
        ):
            # Also reverse order of phases for component a - this should pass
            # and component b should be flagged as missing
            m.params = DummyParameterBlock(
                components={
                    "a": {"phase_equilibrium_form": {("p2", "p1"): "foo"}},
                    "b": {"phase_equilibrium_form": {(1, 2): "foo"}},
                    "c": {},
                },
                phases={
                    "p1": {"equation_of_state": "foo"},
                    "p2": {"equation_of_state": "bar"},
                },
                state_definition="baz",
                pressure_ref=100000.0,
                temperature_ref=300,
                phases_in_equilibrium=[("p1", "p2")],
                phase_equilibrium_state={("p1", "p2"): "whoop"},
                base_units=base_units,
            )

    @pytest.mark.unit
    def test_phases_in_equilibrium_no_formulation(self):
        m = ConcreteModel()

        with pytest.raises(
            ConfigurationError,
            match="params Generic Property Package provided "
            "with a phases_in_equilibrium argument but no "
            "method was specified for "
            "phase_equilibrium_state.",
        ):
            m.params = DummyParameterBlock(
                components={
                    "a": {"phase_equilibrium_form": {("p1", "p2"): "foo"}},
                    "b": {"phase_equilibrium_form": {("p1", "p2"): "foo"}},
                    "c": {"phase_equilibrium_form": {("p1", "p2"): "foo"}},
                },
                phases={
                    "p1": {"equation_of_state": "foo"},
                    "p2": {"equation_of_state": "bar"},
                },
                state_definition="baz",
                pressure_ref=100000.0,
                temperature_ref=300,
                phases_in_equilibrium=[("p1", "p2")],
                base_units=base_units,
            )

    @pytest.mark.unit
    def test_phases_in_equilibrium_missing_pair_formulation(self):
        m = ConcreteModel()

        with pytest.raises(
            ConfigurationError,
            match="params Generic Property Package provided "
            "with a phases_in_equilibrium argument but "
            "phase_equilibrium_state was not specified "
            "for all phase pairs.",
        ):
            # Also reverse order of phases for component a - this should pass
            # and component b should be flagged as missing
            m.params = DummyParameterBlock(
                components={
                    "a": {"phase_equilibrium_form": {("p1", "p2"): "foo"}},
                    "b": {"phase_equilibrium_form": {("p1", "p2"): "foo"}},
                    "c": {"phase_equilibrium_form": {("p1", "p2"): "foo"}},
                },
                phases={
                    "p1": {"equation_of_state": "foo"},
                    "p2": {"equation_of_state": "bar"},
                },
                state_definition="baz",
                pressure_ref=100000.0,
                temperature_ref=300,
                phases_in_equilibrium=[("p1", "p2")],
                phase_equilibrium_state={(1, 2): "whoop"},
                base_units=base_units,
            )

    @pytest.mark.unit
    def test_parameter_construction_no_value(self):
        m = ConcreteModel()

        class test_class:
            # Mook up property method class for testing
            def build_parameters(c):
                c.test_var = Var()

        with pytest.raises(
            ConfigurationError,
            match="params parameter test_var was not assigned "
            "a value. Please check your configuration "
            "arguments.",
        ):
            m.params = DummyParameterBlock(
                components={"a": {"dens_mol_liq_comp": test_class}, "b": {}, "c": {}},
                phases={
                    "p1": {"equation_of_state": DummyEoS},
                    "p2": {"equation_of_state": DummyEoS},
                },
                state_definition=modules[__name__],
                pressure_ref=100000.0,
                temperature_ref=300,
                base_units=base_units,
            )

    @pytest.mark.unit
    def test_parameter_construction_no_data(self):
        m = ConcreteModel()

        class test_class:
            # Mook up property method class for testing
            def build_parameters(c):
                c.config.parameter_data["test"]

        with pytest.raises(
            ConfigurationError,
            match="params values were not defined for "
            "parameter dens_mol_liq_comp in component a. "
            "Please check the parameter_data argument to "
            "ensure values are provided.",
        ):
            m.params = DummyParameterBlock(
                components={"a": {"dens_mol_liq_comp": test_class}, "b": {}, "c": {}},
                phases={
                    "p1": {"equation_of_state": "foo"},
                    "p2": {"equation_of_state": "bar"},
                },
                state_definition=modules[__name__],
                pressure_ref=100000.0,
                temperature_ref=300,
                base_units=base_units,
            )

    @pytest.mark.unit
    def test_no_elements(self):
        m = ConcreteModel()
        m.params = DummyParameterBlock(
            components={"a": {}, "b": {}, "c": {}},
            phases={
                "p1": {
                    "type": LiquidPhase,
                    "component_list": ["a", "b"],
                    "equation_of_state": DummyEoS,
                },
                "p2": {"equation_of_state": DummyEoS},
            },
            state_definition=modules[__name__],
            pressure_ref=100000.0,
            temperature_ref=300,
            base_units=base_units,
        )

        assert not hasattr(m.params, "element_list")
        assert not hasattr(m.params, "element_comp")

    @pytest.mark.unit
    def test_partial_elements(self):
        m = ConcreteModel()

        with pytest.raises(
            ConfigurationError,
            match="params not all Components declared an "
            "elemental_composition argument. Either all "
            "Components must declare this, or none.",
        ):
            m.params = DummyParameterBlock(
                components={
                    "a": {"elemental_composition": {"e1": 1}},
                    "b": {},
                    "c": {},
                },
                phases={
                    "p1": {
                        "type": LiquidPhase,
                        "component_list": ["a", "b"],
                        "equation_of_state": DummyEoS,
                    },
                    "p2": {"equation_of_state": DummyEoS},
                },
                state_definition=modules[__name__],
                pressure_ref=100000.0,
                temperature_ref=300,
                base_units=base_units,
            )

    @pytest.mark.unit
    def test_elements_not_float(self):
        m = ConcreteModel()

        with pytest.raises(
            ConfigurationError,
            match=re.escape(
                "params values in elemental_composition must "
                "be integers (not floats): e1: 2.0."
            ),
        ):
            m.params = DummyParameterBlock(
                components={
                    "a": {"elemental_composition": {"e1": 2.0}},
                    "b": {},
                    "c": {},
                },
                phases={
                    "p1": {
                        "type": LiquidPhase,
                        "component_list": ["a", "b"],
                        "equation_of_state": DummyEoS,
                    },
                    "p2": {"equation_of_state": DummyEoS},
                },
                state_definition=modules[__name__],
                pressure_ref=100000.0,
                temperature_ref=300,
                base_units=base_units,
            )

    @pytest.mark.unit
    def test_elements(self):
        m = ConcreteModel()

        m.params = DummyParameterBlock(
            components={
                "a": {"elemental_composition": {"e1": 1, "e2": 2}},
                "b": {"elemental_composition": {"e3": 3, "e4": 4}},
                "c": {"elemental_composition": {"e1": 5, "e3": 6}},
            },
            phases={
                "p1": {
                    "type": LiquidPhase,
                    "component_list": ["a", "b"],
                    "equation_of_state": DummyEoS,
                },
                "p2": {"equation_of_state": DummyEoS},
            },
            state_definition=modules[__name__],
            pressure_ref=100000.0,
            temperature_ref=300,
            base_units=base_units,
        )

        assert isinstance(m.params.element_list, Set)
        assert len(m.params.element_list) == 4
        assert m.params.element_comp == {
            "a": {"e1": 1, "e2": 2, "e3": 0, "e4": 0},
            "b": {"e1": 0, "e2": 0, "e3": 3, "e4": 4},
            "c": {"e1": 5, "e2": 0, "e3": 6, "e4": 0},
        }

    @pytest.mark.unit
    def test_henry(self):
        m = ConcreteModel()

        m.params = DummyParameterBlock(
            components={
                "a": {
                    "henry_component": {
                        "p1": {"method": modules[__name__], "type": HenryType.Kpx}
                    }
                },
                "b": {},
                "c": {},
            },
            phases={
                "p1": {
                    "type": LiquidPhase,
                    "component_list": ["a", "b"],
                    "equation_of_state": DummyEoS,
                },
                "p2": {"equation_of_state": DummyEoS},
            },
            state_definition=modules[__name__],
            pressure_ref=100000.0,
            temperature_ref=300,
            base_units=base_units,
        )

        assert isinstance(m.params.a.test_param_p1, Var)
        assert m.params.a.test_param_p1.value == 42

    @pytest.mark.unit
    def test_henry_invalid_phase_name(self):
        m = ConcreteModel()

        with pytest.raises(
            ConfigurationError,
            match="params component a was marked as a Henry's "
            "Law component in phase p3, but this is not a "
            "valid phase name.",
        ):
            m.params = DummyParameterBlock(
                components={
                    "a": {"henry_component": {"p3": modules[__name__]}},
                    "b": {},
                    "c": {},
                },
                phases={
                    "p1": {
                        "type": LiquidPhase,
                        "component_list": ["a", "b"],
                        "equation_of_state": DummyEoS,
                    },
                    "p2": {"equation_of_state": DummyEoS},
                },
                state_definition=modules[__name__],
                pressure_ref=100000.0,
                temperature_ref=300,
                base_units=base_units,
            )

    @pytest.mark.unit
    def test_henry_invalid_phase_type(self):
        m = ConcreteModel()

        with pytest.raises(
            ConfigurationError,
            match="params component a was marked as a Henry's "
            "Law component in phase p2, but this is not a "
            "Liquid phase.",
        ):
            m.params = DummyParameterBlock(
                components={
                    "a": {"henry_component": {"p2": modules[__name__]}},
                    "b": {},
                    "c": {},
                },
                phases={
                    "p1": {
                        "type": LiquidPhase,
                        "component_list": ["a", "b"],
                        "equation_of_state": DummyEoS,
                    },
                    "p2": {"equation_of_state": DummyEoS},
                },
                state_definition=modules[__name__],
                pressure_ref=100000.0,
                temperature_ref=300,
                base_units=base_units,
            )

    @pytest.mark.unit
    def test_inherent_reactions(self):
        m = ConcreteModel()
        m.params = DummyParameterBlock(
            components={"a": {}, "b": {}, "c": {}},
            phases={
                "p1": {
                    "type": LiquidPhase,
                    "component_list": ["a", "b"],
                    "equation_of_state": DummyEoS,
                },
                "p2": {"equation_of_state": DummyEoS},
            },
            state_definition=modules[__name__],
            pressure_ref=100000.0,
            temperature_ref=300,
            base_units=base_units,
            inherent_reactions={
                "e1": {
                    "stoichiometry": {("p1", "a"): -3, ("p1", "b"): 4},
                    "heat_of_reaction": "foo",
                    "equilibrium_form": "foo",
                }
            },
        )

        assert m.params.has_inherent_reactions

        assert isinstance(m.params.inherent_reaction_idx, Set)
        assert len(m.params.inherent_reaction_idx) == 1
        assert m.params.inherent_reaction_idx == ["e1"]

        assert hasattr(m.params, "inherent_reaction_stoichiometry")
        assert len(m.params.inherent_reaction_stoichiometry) == 5
        assert m.params.inherent_reaction_stoichiometry == {
            ("e1", "p1", "a"): -3,
            ("e1", "p1", "b"): 4,
            ("e1", "p2", "a"): 0,
            ("e1", "p2", "b"): 0,
            ("e1", "p2", "c"): 0,
        }

        assert isinstance(m.params.reaction_e1, Block)
        assert isinstance(m.params.reaction_e1.reaction_order, Var)
        assert len(m.params.reaction_e1.reaction_order) == 5
        for i in m.params.reaction_e1.reaction_order:
            order = {
                ("p1", "a"): -3,
                ("p1", "b"): 4,
                ("p2", "a"): 0,
                ("p2", "b"): 0,
                ("p2", "c"): 0,
            }
            assert m.params.reaction_e1.reaction_order[i].value == order[i]

    @pytest.mark.unit
    def test_inherent_reactions_no_stoichiometry(self):
        m = ConcreteModel()

        with pytest.raises(
            ConfigurationError,
            match="params inherent reaction e1 was not "
            "provided with a stoichiometry configuration "
            "argument.",
        ):
            m.params = DummyParameterBlock(
                components={"a": {}, "b": {}, "c": {}},
                phases={
                    "p1": {
                        "type": LiquidPhase,
                        "component_list": ["a", "b"],
                        "equation_of_state": DummyEoS,
                    },
                    "p2": {"equation_of_state": DummyEoS},
                },
                state_definition=modules[__name__],
                pressure_ref=100000.0,
                temperature_ref=300,
                base_units=base_units,
                inherent_reactions={
                    "e1": {"heat_of_reaction": "foo", "equilibrium_form": "foo"}
                },
            )

    @pytest.mark.unit
    def test_inherent_reactions_invalid_phase_stoichiometry(self):
        m = ConcreteModel()

        with pytest.raises(
            ConfigurationError,
            match="params stoichiometry for inherent "
            "reaction e1 included unrecognised phase p7.",
        ):
            m.params = DummyParameterBlock(
                components={"a": {}, "b": {}, "c": {}},
                phases={
                    "p1": {
                        "type": LiquidPhase,
                        "component_list": ["a", "b"],
                        "equation_of_state": DummyEoS,
                    },
                    "p2": {"equation_of_state": DummyEoS},
                },
                state_definition=modules[__name__],
                pressure_ref=100000.0,
                temperature_ref=300,
                base_units=base_units,
                inherent_reactions={
                    "e1": {
                        "stoichiometry": {("p7", "a"): -3, ("p1", "b"): 4},
                        "heat_of_reaction": "foo",
                        "equilibrium_form": "foo",
                    }
                },
            )

    @pytest.mark.unit
    def test_inherent_reactions_invalid_component_stoichiometry(self):
        m = ConcreteModel()

        with pytest.raises(
            ConfigurationError,
            match="params stoichiometry for inherent "
            "reaction e1 included unrecognised component c7.",
        ):
            m.params = DummyParameterBlock(
                components={"a": {}, "b": {}, "c": {}},
                phases={
                    "p1": {
                        "type": LiquidPhase,
                        "component_list": ["a", "b"],
                        "equation_of_state": DummyEoS,
                    },
                    "p2": {"equation_of_state": DummyEoS},
                },
                state_definition=modules[__name__],
                pressure_ref=100000.0,
                temperature_ref=300,
                base_units=base_units,
                inherent_reactions={
                    "e1": {
                        "stoichiometry": {("p1", "c7"): -3, ("p1", "b"): 4},
                        "heat_of_reaction": "foo",
                        "equilibrium_form": "foo",
                    }
                },
            )

    @pytest.mark.unit
    def test_inherent_reactions_no_form(self):
        m = ConcreteModel()

        with pytest.raises(
            ConfigurationError,
            match="params inherent reaction e1 was not "
            "provided with a equilibrium_form configuration "
            "argument.",
        ):
            m.params = DummyParameterBlock(
                components={"a": {}, "b": {}, "c": {}},
                phases={
                    "p1": {
                        "type": LiquidPhase,
                        "component_list": ["a", "b"],
                        "equation_of_state": DummyEoS,
                    },
                    "p2": {"equation_of_state": DummyEoS},
                },
                state_definition=modules[__name__],
                pressure_ref=100000.0,
                temperature_ref=300,
                base_units=base_units,
                inherent_reactions={
                    "e1": {"stoichiometry": {("p1", "a"): -3, ("p1", "b"): 4}}
                },
            )

    @pytest.mark.unit
    def test_default_scaling(self):
        m = ConcreteModel()
        m.params = DummyParameterBlock(
            components={"a": {}, "b": {}, "c": {}},
            phases={
                "p1": {
                    "type": LiquidPhase,
                    "component_list": ["a", "b"],
                    "equation_of_state": DummyEoS,
                },
                "p2": {"equation_of_state": DummyEoS},
            },
            state_definition=modules[__name__],
            pressure_ref=100000.0,
            temperature_ref=300,
            base_units=base_units,
        )

        dsf = m.params.default_scaling_factor

        assert dsf[("temperature", None)] == 1e-2
        assert dsf[("pressure", None)] == 1e-5
        assert dsf[("dens_mol_phase", None)] == 1e-2
        assert dsf[("enth_mol", None)] == 1e-4
        assert dsf[("entr_mol", None)] == 1e-2
        assert dsf[("fug_phase_comp", None)] == 1e-4
        assert dsf[("fug_coeff_phase_comp", None)] == 1
        assert dsf[("gibbs_mol", None)] == 1e-4
        assert dsf[("mole_frac_comp", None)] == 1e3
        assert dsf[("mole_frac_phase_comp", None)] == 1e3
        assert dsf[("mw", None)] == 1e3
        assert dsf[("mw_comp", None)] == 1e3
        assert dsf[("mw_phase", None)] == 1e3

    @pytest.mark.unit
    def test_default_scaling_convert(self):
        m = ConcreteModel()
        m.params = DummyParameterBlock(
            components={"a": {}, "b": {}, "c": {}},
            phases={
                "p1": {
                    "type": LiquidPhase,
                    "component_list": ["a", "b"],
                    "equation_of_state": DummyEoS,
                },
                "p2": {"equation_of_state": DummyEoS},
            },
            state_definition=modules[__name__],
            pressure_ref=100000.0,
            temperature_ref=300,
            base_units={
                "time": pyunits.minutes,
                "length": pyunits.foot,
                "mass": pyunits.pound,
                "amount": pyunits.mol,
                "temperature": pyunits.degR,
            },
        )

        dsf = m.params.default_scaling_factor

        assert dsf[("temperature", None)] == 1e-2
        assert dsf[("pressure", None)] == 1e-8
        assert dsf[("dens_mol_phase", None)] == 1
        assert dsf[("enth_mol", None)] == 1e-9
        assert dsf[("entr_mol", None)] == 1e-7
        assert dsf[("fug_phase_comp", None)] == 1e-7
        assert dsf[("fug_coeff_phase_comp", None)] == 1
        assert dsf[("gibbs_mol", None)] == 1e-9
        assert dsf[("mole_frac_comp", None)] == 1e3
        assert dsf[("mole_frac_phase_comp", None)] == 1e3
        assert dsf[("mw", None)] == 1e3
        assert dsf[("mw_comp", None)] == 1e3
        assert dsf[("mw_phase", None)] == 1e3


# -----------------------------------------------------------------------------
# Dummy methods for testing build calls to sub-modules
def define_state(b):
    b.state_defined = True
    b.temperature = Var(initialize=100)


def get_material_flow_basis(self, *args, **kwargs):
    return MaterialFlowBasis.molar


def dummy_method(*args, **kwargs):
    return 42


class TestGenericStateBlock(object):
    @pytest.fixture()
    def frame(self):
        m = ConcreteModel()
        m.params = DummyParameterBlock(
            components={
                "a": {
                    "type": Solvent,
                    "diffus_phase_comp": {"p1": dummy_method, "p2": dummy_method},
                    "visc_d_phase_comp": {"p1": dummy_method, "p2": dummy_method},
                    "therm_cond_phase_comp": {"p1": dummy_method, "p2": dummy_method},
                    "pressure_sat_comp": dummy_method,
                },
                "b": {
                    "diffus_phase_comp": {"p1": dummy_method, "p2": dummy_method},
                    "visc_d_phase_comp": {"p1": dummy_method, "p2": dummy_method},
                    "therm_cond_phase_comp": {"p1": dummy_method, "p2": dummy_method},
                    "pressure_sat_comp": dummy_method,
                },
                "c": {
                    "diffus_phase_comp": {"p1": dummy_method, "p2": dummy_method},
                    "visc_d_phase_comp": {"p1": dummy_method, "p2": dummy_method},
                    "therm_cond_phase_comp": {"p1": dummy_method, "p2": dummy_method},
                    "pressure_sat_comp": dummy_method,
                },
            },
            phases={
                "p1": {
                    "equation_of_state": DummyEoS,
                    "therm_cond_phase": dummy_method,
                    "surf_tens_phase": dummy_method,
                    "visc_d_phase": dummy_method,
                },
                "p2": {
                    "equation_of_state": DummyEoS,
                    "therm_cond_phase": dummy_method,
                    "surf_tens_phase": dummy_method,
                    "visc_d_phase": dummy_method,
                },
            },
            state_definition=modules[__name__],
            pressure_ref=100000.0,
            temperature_ref=300,
            base_units=base_units,
        )

        m.props = m.params.build_state_block([1], defined_state=False)
        m.props[1].get_material_flow_basis = MethodType(
            get_material_flow_basis, m.props[1].get_material_flow_basis
        )

        # Add necessary variables to state block
        m.props[1].flow_mol = Var(bounds=(0, 3000))
        m.props[1].pressure = Var(bounds=(1000, 3000))
        m.props[1].temperature = Var(bounds=(100, 200))
        m.props[1].mole_frac_phase_comp = Var(
            m.params.phase_list, m.params.component_list, initialize=0.3
        )
        m.props[1].phase_frac = Var(m.params.phase_list)

        return m

    @pytest.mark.unit
    def test_build(self, frame):
        assert isinstance(frame.props, Block)
        assert len(frame.props) == 1
        assert frame.props.default_initializer is BlockTriangularizationInitializer

        # Check for expected behaviour for dummy methods
        assert frame.props[1].state_defined
        assert isinstance(frame.props[1].dummy_var, Var)
        assert frame.props[1].eos_common == 2

    @pytest.mark.unit
    def test_basic_scaling(self, frame):
        frame.props[1].calculate_scaling_factors()

        assert frame.props[1].scaling_check

        assert len(frame.props[1].scaling_factor) == 8
        assert frame.props[1].scaling_factor[frame.props[1].temperature] == 0.01
        assert frame.props[1].scaling_factor[frame.props[1].pressure] == 1e-5
        assert (
            frame.props[1].scaling_factor[
                frame.props[1].mole_frac_phase_comp["p1", "a"]
            ]
            == 1000
        )
        assert (
            frame.props[1].scaling_factor[
                frame.props[1].mole_frac_phase_comp["p1", "b"]
            ]
            == 1000
        )
        assert (
            frame.props[1].scaling_factor[
                frame.props[1].mole_frac_phase_comp["p1", "c"]
            ]
            == 1000
        )
        assert (
            frame.props[1].scaling_factor[
                frame.props[1].mole_frac_phase_comp["p2", "a"]
            ]
            == 1000
        )
        assert (
            frame.props[1].scaling_factor[
                frame.props[1].mole_frac_phase_comp["p2", "b"]
            ]
            == 1000
        )
        assert (
            frame.props[1].scaling_factor[
                frame.props[1].mole_frac_phase_comp["p2", "c"]
            ]
            == 1000
        )

    @pytest.mark.unit
    def test_build_all(self, frame):
        # Add necessary parameters
        frame.params.a.mw = 1
        frame.params.b.mw = 2
        frame.params.c.mw = 3

        # Add variables that would be built by state definition
        frame.props[1].flow_mol_phase_comp = Var(frame.props[1].phase_component_set)

        # Call all properties in metadata and assert they exist.
        for p in frame.params.get_metadata().properties.list_supported_properties():
            if p.name.endswith(("apparent", "true")):
                # True and apparent properties require electrolytes, which are
                # not tested here
                # Check that method exists and continue
                if frame.params.get_metadata().properties[p.name].method is not None:
                    assert hasattr(
                        frame.props[1],
                        frame.params.get_metadata().properties[p.name].method,
                    )
                continue
            elif p.name.endswith(("_crit")):
                # Critical properties will be tested elsewhere
                assert hasattr(
                    frame.props[1],
                    frame.params.get_metadata().properties[p.name].method,
                )
                continue
            elif p.name.endswith(("bubble", "bub", "dew")):
                # Bubble and dew properties require phase equilibria, which are
                # not tested here
                # Check that method exists and continue
                assert hasattr(
                    frame.props[1],
                    frame.params.get_metadata().properties[p.name].method,
                )
                continue
            elif p.name in ["dh_rxn", "log_k_eq"]:
                # Not testing inherent reactions here either
                # Check that method exists and continue
                assert hasattr(
                    frame.props[1],
                    frame.params.get_metadata().properties[p.name].method,
                )
                continue
            elif p.name in {
                "diffus_phase_comp",
            }:
                # phase indexed properties - these will be tested separately.
                continue
            elif p.supported:
                assert hasattr(frame.props[1], p.name)
            else:
                assert not hasattr(frame.props[1], p.name)

    @pytest.mark.unit
    def test_flows(self, frame):

        # Need to set the material flow basis for this test
        def set_flow_basis():
            return MaterialFlowBasis.molar

        frame.props[1].get_material_flow_basis = set_flow_basis

        frame.props[1].flow_mol.fix(100)
        frame.props[1].phase_frac["p1"].fix(0.75)
        frame.props[1].phase_frac["p2"].fix(0.25)
        frame.props[1].mole_frac_phase_comp["p1", "a"].fix(0.1)
        frame.props[1].mole_frac_phase_comp["p1", "b"].fix(0.2)
        frame.props[1].mole_frac_phase_comp["p1", "c"].fix(0.7)
        frame.props[1].mole_frac_phase_comp["p2", "c"].fix(0.1)
        frame.props[1].mole_frac_phase_comp["p2", "b"].fix(0.2)
        frame.props[1].mole_frac_phase_comp["p2", "a"].fix(0.7)
        frame.props[1].params.get_component("a").mw = 2
        frame.props[1].params.get_component("b").mw = 3
        frame.props[1].params.get_component("c").mw = 4

        assert value(frame.props[1].flow_vol) == pytest.approx(100 / 55e3, rel=1e-4)
        assert value(frame.props[1].flow_vol_phase["p1"]) == pytest.approx(
            100 / 55e3 * 0.75, rel=1e-4
        )
        assert value(frame.props[1].flow_vol_phase["p2"]) == pytest.approx(
            100 / 55e3 * 0.25, rel=1e-4
        )
        assert value(frame.props[1].flow_mol_comp["a"]) == pytest.approx(
            100 * 0.1 * 0.75 + 100 * 0.7 * 0.25, rel=1e-4
        )
        assert value(frame.props[1].flow_mol_comp["b"]) == pytest.approx(
            100 * 0.2 * 0.75 + 100 * 0.2 * 0.25, rel=1e-4
        )
        assert value(frame.props[1].flow_mol_comp["c"]) == pytest.approx(
            100 * 0.1 * 0.25 + 100 * 0.7 * 0.75, rel=1e-4
        )
        assert value(frame.props[1].mw_comp["a"]) == pytest.approx(2, rel=1e-4)
        assert value(frame.props[1].mw_phase["p1"]) == pytest.approx(
            2 * 0.1 + 3 * 0.2 + 4 * 0.7, rel=1e-4
        )
        assert value(frame.props[1].mw) == pytest.approx(
            0.75 * (2 * 0.1 + 3 * 0.2 + 4 * 0.7) + 0.25 * (2 * 0.7 + 3 * 0.2 + 4 * 0.1),
            rel=1e-4,
        )
        assert value(frame.props[1].flow_mass) == pytest.approx(
            100 * 0.75 * (2 * 0.1 + 3 * 0.2 + 4 * 0.7)
            + 100 * 0.25 * (2 * 0.7 + 3 * 0.2 + 4 * 0.1),
            rel=1e-4,
        )
        assert value(frame.props[1].flow_mass_phase["p1"]) == pytest.approx(
            100 * 0.75 * (2 * 0.1 + 3 * 0.2 + 4 * 0.7), rel=1e-4
        )
        assert value(frame.props[1].flow_mass_phase["p2"]) == pytest.approx(
            100 * 0.25 * (2 * 0.7 + 3 * 0.2 + 4 * 0.1), rel=1e-4
        )

    class dummy_prop:
        def return_expression(*args, **kwargs):
            return 4

        class visc_d_phase_comp(object):
            def return_expression(*args, **kwargs):
                return 5

        class therm_cond_phase_comp(object):
            def return_expression(*args, **kwargs):
                return 7

    @pytest.mark.unit
    def test_diffus_phase_comp(self, frame):
        frame.params.a.config.diffus_phase_comp = {
            "p1": TestGenericStateBlock.dummy_prop,
            "p2": TestGenericStateBlock.dummy_prop,
        }
        frame.params.b.config.diffus_phase_comp = {
            "p2": TestGenericStateBlock.dummy_prop
        }
        frame.params.c.config.diffus_phase_comp = {
            "p1": TestGenericStateBlock.dummy_prop
        }

        # There should be two skipped indices, so length should be 4
        assert len(frame.props[1].diffus_phase_comp) == 4
        assert value(frame.props[1].diffus_phase_comp["p1", "a"]) == 4
        assert value(frame.props[1].diffus_phase_comp["p2", "a"]) == 4
        assert value(frame.props[1].diffus_phase_comp["p2", "b"]) == 4
        assert value(frame.props[1].diffus_phase_comp["p1", "c"]) == 4

    @pytest.mark.unit
    def test_visc_d_phase_comp(self, frame):
        frame.params.a.config.visc_d_phase_comp = {
            "p1": TestGenericStateBlock.dummy_prop,
            "p2": TestGenericStateBlock.dummy_prop,
        }
        frame.params.b.config.visc_d_phase_comp = {
            "p2": TestGenericStateBlock.dummy_prop
        }
        frame.params.c.config.visc_d_phase_comp = {
            "p1": TestGenericStateBlock.dummy_prop
        }
        # Have to explicitly call constructor
        assert not hasattr(frame.props[1], "_visc_d_phase_comp")
        frame.props[1]._make_visc_d_phase_comp()

        # There should be two skipped indices, so length should be 4
        assert len(frame.props[1]._visc_d_phase_comp) == 4
        assert value(frame.props[1]._visc_d_phase_comp["p1", "a"]) == 5
        assert value(frame.props[1]._visc_d_phase_comp["p2", "a"]) == 5
        assert value(frame.props[1]._visc_d_phase_comp["p2", "b"]) == 5
        assert value(frame.props[1]._visc_d_phase_comp["p1", "c"]) == 5

    @pytest.mark.unit
    def test_therm_cond_phase_comp(self, frame):
        frame.params.a.config.therm_cond_phase_comp = {
            "p1": TestGenericStateBlock.dummy_prop,
            "p2": TestGenericStateBlock.dummy_prop,
        }
        frame.params.b.config.therm_cond_phase_comp = {
            "p2": TestGenericStateBlock.dummy_prop
        }
        frame.params.c.config.therm_cond_phase_comp = {
            "p1": TestGenericStateBlock.dummy_prop
        }
        # Have to explicitly call constructor
        assert not hasattr(frame.props[1], "_therm_cond_phase_comp")
        frame.props[1]._make_therm_cond_phase_comp()

        # There should be two skipped indices, so length should be 4
        assert len(frame.props[1]._therm_cond_phase_comp) == 4
        assert value(frame.props[1]._therm_cond_phase_comp["p1", "a"]) == 7
        assert value(frame.props[1]._therm_cond_phase_comp["p2", "a"]) == 7
        assert value(frame.props[1]._therm_cond_phase_comp["p2", "b"]) == 7
        assert value(frame.props[1]._therm_cond_phase_comp["p1", "c"]) == 7

    @pytest.mark.unit
    def test_surf_tens_phase(self, frame):
        # Surface tension is only built for Liquid Phases, and we have none.
        # All values should be None.
        for p in frame.props[1].phase_list:
            assert p not in frame.props[1].surf_tens_phase

    @pytest.mark.unit
    def test_therm_cond_phase(self, frame):
        frame.params.p1.config.therm_cond_phase = TestGenericStateBlock.dummy_prop
        frame.params.p2.config.therm_cond_phase = TestGenericStateBlock.dummy_prop

        for p in frame.props[1].phase_list:
            assert value(frame.props[1].therm_cond_phase[p]) == 4

    @pytest.mark.unit
    def test_visc_d_phase(self, frame):
        frame.params.p1.config.visc_d_phase = TestGenericStateBlock.dummy_prop
        frame.params.p2.config.visc_d_phase = TestGenericStateBlock.dummy_prop

        for p in frame.props[1].phase_list:
            assert value(frame.props[1].visc_d_phase[p]) == 4


class TestCriticalProps:
    @pytest.mark.unit
    def test_get_critical_ref_phase_exception(self):
        # No vapor or liquid phases
        m = ConcreteModel()
        m.params = DummyParameterBlock(
            components={
                "a": {},
            },
            phases={
                "p1": {
                    "equation_of_state": DummyEoS,
                },
                "p2": {
                    "equation_of_state": DummyEoS,
                },
            },
            state_definition=modules[__name__],
            pressure_ref=100000.0,
            temperature_ref=300,
            base_units=base_units,
        )

        m.props = m.params.build_state_block([1], defined_state=False)

        with pytest.raises(
            AttributeError,
            match="No liquid or vapor phase found to use as reference phase "
            "for calculating critical properties.",
        ):
            m.props[1]._get_critical_ref_phase()

    @pytest.mark.unit
    def test_get_critical_ref_phase_exception(self):
        # No vapor or liquid phases
        m = ConcreteModel()
        m.params = DummyParameterBlock(
            components={
                "a": {},
            },
            phases={
                "p1": {
                    "equation_of_state": DummyEoS,
                },
                "p2": {
                    "equation_of_state": DummyEoS,
                },
            },
            state_definition=modules[__name__],
            pressure_ref=100000.0,
            temperature_ref=300,
            base_units=base_units,
        )

        m.props = m.params.build_state_block([1], defined_state=False)

        with pytest.raises(
            PropertyPackageError,
            match="No liquid or vapor phase found to use as reference phase "
            "for calculating critical properties.",
        ):
            m.props[1]._get_critical_ref_phase()

    @pytest.mark.unit
    def test_get_critical_ref_phase_VL(self):
        # No vapor or liquid phases
        m = ConcreteModel()
        m.params = DummyParameterBlock(
            components={
                "a": {},
            },
            phases={
                "p1": {
                    "type": LiquidPhase,
                    "equation_of_state": DummyEoS,
                },
                "p2": {
                    "type": VaporPhase,
                    "equation_of_state": DummyEoS,
                },
            },
            state_definition=modules[__name__],
            pressure_ref=100000.0,
            temperature_ref=300,
            base_units=base_units,
        )

        m.props = m.params.build_state_block([1], defined_state=False)

        assert m.props[1]._get_critical_ref_phase() == "p1"

    @pytest.mark.unit
    def test_get_critical_ref_phase_LL(self):
        # No vapor or liquid phases
        m = ConcreteModel()
        m.params = DummyParameterBlock(
            components={
                "a": {},
            },
            phases={
                "p1": {
                    "type": LiquidPhase,
                    "equation_of_state": DummyEoS,
                },
                "p2": {
                    "type": LiquidPhase,
                    "equation_of_state": DummyEoS,
                },
            },
            state_definition=modules[__name__],
            pressure_ref=100000.0,
            temperature_ref=300,
            base_units=base_units,
        )

        m.props = m.params.build_state_block([1], defined_state=False)

        assert m.props[1]._get_critical_ref_phase() == "p1"

    @pytest.mark.unit
    def test_get_critical_ref_phase_VLL(self):
        class DummyPE:
            def phase_equil(self, *args, **kwargs):
                pass

            @staticmethod
            def return_expression(b, *args, **kwargs):
                return b.temperature == 42

        # No vapor or liquid phases
        m = ConcreteModel()
        m.params = DummyParameterBlock(
            components={
                "a": {
                    "phase_equilibrium_form": {("p2", "p3"): DummyPE},
                },
            },
            phases={
                "p1": {
                    "type": LiquidPhase,
                    "equation_of_state": DummyEoS,
                },
                "p2": {
                    "type": VaporPhase,
                    "equation_of_state": DummyEoS,
                },
                "p3": {
                    "type": LiquidPhase,
                    "equation_of_state": DummyEoS,
                },
            },
            state_definition=modules[__name__],
            pressure_ref=100000.0,
            temperature_ref=300,
            base_units=base_units,
            phases_in_equilibrium=[("p2", "p3")],
            phase_equilibrium_state={("p2", "p3"): DummyPE},
        )

        m.props = m.params.build_state_block([1], defined_state=False)

        assert m.props[1]._get_critical_ref_phase() == "p3"

    @pytest.mark.unit
    def test_get_critical_ref_phase_VX(self):
        # No vapor or liquid phases
        m = ConcreteModel()
        m.params = DummyParameterBlock(
            components={
                "a": {},
            },
            phases={
                "p1": {
                    "type": VaporPhase,
                    "equation_of_state": DummyEoS,
                },
                "p2": {
                    "equation_of_state": DummyEoS,
                },
            },
            state_definition=modules[__name__],
            pressure_ref=100000.0,
            temperature_ref=300,
            base_units=base_units,
        )

        m.props = m.params.build_state_block([1], defined_state=False)

        assert m.props[1]._get_critical_ref_phase() == "p1"

    @pytest.mark.unit
    def test_critical_props_not_implemented(self):
        m = ConcreteModel()
        m.params = DummyParameterBlock(
            components={
                "a": {},
            },
            phases={
                "p1": {
                    "type": LiquidPhase,
                    "equation_of_state": DummyEoS,
                },
            },
            state_definition=modules[__name__],
            pressure_ref=100000.0,
            temperature_ref=300,
            base_units=base_units,
        )

        m.props = m.params.build_state_block([1], defined_state=False)

        with pytest.raises(
            NotImplementedError,
            match=re.escape(
                "props[1] Equation of State module has not implemented a method for "
                "build_critical_properties. Please contact the EoS developer or use a "
                "different module."
            ),
        ):
            m.props[1]._critical_props()

    @pytest.mark.unit
    def test_critical_props_dummy_method(self):
        class DummyEoS2(DummyEoS):
            @staticmethod
            def build_critical_properties(b, *args, **kwargs):
                b._dummy_crit_executed = True

        m = ConcreteModel()
        m.params = DummyParameterBlock(
            components={
                "a": {},
            },
            phases={
                "p1": {
                    "type": LiquidPhase,
                    "equation_of_state": DummyEoS2,
                },
            },
            state_definition=modules[__name__],
            pressure_ref=100000.0,
            temperature_ref=300,
            base_units=base_units,
        )

        m.props = m.params.build_state_block([1], defined_state=False)
        m.props[1]._critical_props()

        assert m.props[1]._dummy_crit_executed

        assert isinstance(m.props[1].compress_fact_crit, Var)
        assert isinstance(m.props[1].dens_mol_crit, Var)
        assert isinstance(m.props[1].pressure_crit, Var)
        assert isinstance(m.props[1].temperature_crit, Var)

    @pytest.mark.unit
    def test_critical_props_attribute_error(self):
        class DummyEoS2(DummyEoS):
            @staticmethod
            def build_critical_properties(b, *args, **kwargs):
                raise AttributeError()

        m = ConcreteModel()
        m.params = DummyParameterBlock(
            components={
                "a": {},
            },
            phases={
                "p1": {
                    "type": LiquidPhase,
                    "equation_of_state": DummyEoS2,
                },
            },
            state_definition=modules[__name__],
            pressure_ref=100000.0,
            temperature_ref=300,
            base_units=base_units,
        )

        m.props = m.params.build_state_block([1], defined_state=False)
        try:
            m.props[1]._critical_props()
        except AttributeError:
            assert not hasattr(m.props[1], "compress_fact_crit")
            assert not hasattr(m.props[1], "dens_mol_crit")
            assert not hasattr(m.props[1], "pressure_crit")
            assert not hasattr(m.props[1], "temperature_crit")

    @pytest.mark.unit
    def test_initialize_critical_props(self):
        m = ConcreteModel()

        class DummyEoS2(DummyEoS):
            @staticmethod
            def build_critical_properties(b, *args, **kwargs):
                b._dummy_crit_executed = True

        # Dummy params block
        m.params = DummyParameterBlock(
            components={
                "a": {
                    "parameter_data": {
                        "compress_fact_crit": 0.1,
                        "dens_mol_crit": 1,
                        "pressure_crit": 1e5,
                        "temperature_crit": 100,
                    }
                },
                "b": {
                    "parameter_data": {
                        "compress_fact_crit": 0.2,
                        "dens_mol_crit": 2,
                        "pressure_crit": 2e5,
                        "temperature_crit": 200,
                    }
                },
                "c": {
                    "parameter_data": {
                        "compress_fact_crit": 0.3,
                        "dens_mol_crit": 3,
                        "pressure_crit": 3e5,
                        "temperature_crit": 300,
                    }
                },
            },
            phases={
                "Vap": {
                    "type": LiquidPhase,
                    "equation_of_state": DummyEoS2,
                },
                "Liq": {
                    "type": LiquidPhase,
                    "equation_of_state": DummyEoS2,
                },
            },
            base_units={
                "time": pyunits.s,
                "length": pyunits.m,
                "mass": pyunits.kg,
                "amount": pyunits.mol,
                "temperature": pyunits.K,
            },
            state_definition=modules[__name__],
            pressure_ref=100000.0,
            temperature_ref=300,
            parameter_data={
                "PR_kappa": {
                    ("a", "a"): 0.0,
                    ("a", "b"): 0.0,
                    ("a", "c"): 0.0,
                    ("b", "a"): 0.0,
                    ("b", "b"): 0.0,
                    ("b", "c"): 0.0,
                    ("c", "a"): 0.0,
                    ("c", "b"): 0.0,
                    ("c", "c"): 0.0,
                },
            },
        )

        m.props = m.params.build_state_block([1], defined_state=False)

        # Add common variables
        m.props[1].mole_frac_comp = Var(
            m.params.component_list, initialize=0.5, bounds=(1e-12, 1)
        )
        m.props[1].mole_frac_comp["a"].fix(0.4)
        m.props[1].mole_frac_comp["b"].fix(0.6)
        m.props[1].mole_frac_comp["c"].fix(1e-8)

        # Build critical props
        m.props[1]._critical_props()

        # Initialize critical props
        _initialize_critical_props(m.props[1])

        # these equalities are expected to be as exact as possible,
        # i.e. very small tolerances
        almost_exactly = functools.partial(pytest.approx, rel=1e-12)
        assert value(m.props[1].compress_fact_crit) == almost_exactly(
            0.4 * 0.1 + 0.6 * 0.2 + 1e-8 * 0.3
        )
        assert value(m.props[1].dens_mol_crit) == almost_exactly(
            0.4 * 1 + 0.6 * 2 + 1e-8 * 3
        )
        assert value(m.props[1].pressure_crit) == almost_exactly(
            0.4 * 1e5 + 0.6 * 2e5 + 1e-8 * 3e5
        )
        assert value(m.props[1].temperature_crit) == almost_exactly(
            0.4 * 1e2 + 0.6 * 2e2 + 1e-8 * 3e2
        )

    @pytest.mark.unit
    def test_initialize_critical_props_missing_value(self):
        m = ConcreteModel()

        class DummyEoS2(DummyEoS):
            @staticmethod
            def build_critical_properties(b, *args, **kwargs):
                b._dummy_crit_executed = True

        # Dummy params block
        m.params = DummyParameterBlock(
            components={
                "a": {
                    "parameter_data": {
                        "dens_mol_crit": 1,  # Missing Z_crit
                        "pressure_crit": 1e5,
                        "temperature_crit": 100,
                    }
                },
                "b": {
                    "parameter_data": {
                        "compress_fact_crit": 0.2,
                        "dens_mol_crit": 2,
                        "pressure_crit": 2e5,
                        "temperature_crit": 200,
                    }
                },
                "c": {
                    "parameter_data": {
                        "compress_fact_crit": 0.3,
                        "dens_mol_crit": 3,
                        "pressure_crit": 3e5,
                        "temperature_crit": 300,
                    }
                },
            },
            phases={
                "Vap": {
                    "type": LiquidPhase,
                    "equation_of_state": DummyEoS2,
                },
                "Liq": {
                    "type": LiquidPhase,
                    "equation_of_state": DummyEoS2,
                },
            },
            base_units={
                "time": pyunits.s,
                "length": pyunits.m,
                "mass": pyunits.kg,
                "amount": pyunits.mol,
                "temperature": pyunits.K,
            },
            state_definition=modules[__name__],
            pressure_ref=100000.0,
            temperature_ref=300,
            parameter_data={
                "PR_kappa": {
                    ("a", "a"): 0.0,
                    ("a", "b"): 0.0,
                    ("a", "c"): 0.0,
                    ("b", "a"): 0.0,
                    ("b", "b"): 0.0,
                    ("b", "c"): 0.0,
                    ("c", "a"): 0.0,
                    ("c", "b"): 0.0,
                    ("c", "c"): 0.0,
                },
            },
        )

        m.props = m.params.build_state_block([1], defined_state=False)

        # Add common variables
        m.props[1].mole_frac_comp = Var(
            m.params.component_list, initialize=0.5, bounds=(1e-12, 1)
        )
        m.props[1].mole_frac_comp["a"].fix(0.4)
        m.props[1].mole_frac_comp["b"].fix(0.6)
        m.props[1].mole_frac_comp["c"].fix(1e-8)

        # Build critical props
        m.props[1]._critical_props()

        # Initialize critical props
        with pytest.raises(
            AttributeError,
            match="Missing attribute found when initializing compress_fact_crit. "
            "Make sure you have provided values for compress_fact_crit in all "
            "Component declarations.",
        ):
            _initialize_critical_props(m.props[1])


# Invalid property configuration to trigger configuration error
configuration = {
    # Specifying components
    "components": {
        "H2O": {
            "type": Component,
            "parameter_data": {
                "pressure_crit": (220.6e5, pyunits.Pa),
                "temperature_crit": (647, pyunits.K),
                "omega": 0.344,
            },
        },
    },
    # Specifying phases
    "phases": {
        "Liq": {
            "type": LiquidPhase,
            "equation_of_state": Cubic,
            "equation_of_state_options": {"type": CubicType.PR},
        },
        "Vap": {
            "type": VaporPhase,
            "equation_of_state": Cubic,
            "equation_of_state_options": {"type": CubicType.PR},
        },
    },
    # Set base units of measurement
    "base_units": {
        "time": pyunits.s,
        "length": pyunits.m,
        "mass": pyunits.kg,
        "amount": pyunits.mol,
        "temperature": pyunits.K,
    },
    # Specifying state definition
    "state_definition": FTPx,
    "state_bounds": {
        "flow_mol": (0, 100, 1000, pyunits.mol / pyunits.s),
        "temperature": (273.15, 300, 500, pyunits.K),
        "pressure": (5e4, 1e5, 1e6, pyunits.Pa),
    },
    "pressure_ref": (101325, pyunits.Pa),
    "temperature_ref": (298.15, pyunits.K),
    "parameter_data": {
        "PR_kappa": {
            ("foo", "bar"): 0.000,
        }
    },
}


@pytest.mark.integration
def test_phase_component_flash():
    # Regression test for issue #1423
    m = ConcreteModel()
    m.fs = FlowsheetBlock(dynamic=False)

    m.fs.props = GenericParameterBlock(**BTconfig)

    m.fs.flash = Flash(
        property_package=m.fs.props,
        material_balance_type=MaterialBalanceType.componentPhase,
    )

    assert m.fs.props.phase_equilibrium_list == {
        "PE1": ["benzene", ("Vap", "Liq")],
        "PE2": ["toluene", ("Vap", "Liq")],
    }
    assert isinstance(m.fs.flash, Flash)



#################################################################################
# The Institute for the Design of Advanced Energy Systems Integrated Platform
# Framework (IDAES IP) was produced under the DOE Institute for the
# Design of Advanced Energy Systems (IDAES).
#
# Copyright (c) 2018-2024 by the software owners: The Regents of the
# University of California, through Lawrence Berkeley National Laboratory,
# National Technology & Engineering Solutions of Sandia, LLC, Carnegie Mellon
# University, West Virginia University Research Corporation, et al.
# All rights reserved.  Please see the files COPYRIGHT.md and LICENSE.md
# for full copyright and license information.
#################################################################################
"""
Make sure the subcritical boiler recirculation system example solves.
"""

__author__ = "Miguel Zamarripa"

import pytest
import pyomo.environ as pyo
from pyomo.util.check_units import assert_units_consistent

from idaes.models_extra.power_generation.flowsheets.subcritical_power_plant.subcritical_boiler import (
    main,
)
from idaes.core.util.model_statistics import (
    degrees_of_freedom,
    activated_equalities_generator,
)
from idaes.models.properties import iapws95
from idaes.core.solvers import get_solver
import idaes.core.util.scaling as iscale
from idaes.models.properties.general_helmholtz import helmholtz_available

solver = get_solver()


pytestmark = pytest.mark.usefixtures("run_module_in_tmp_path")


@pytest.fixture(scope="module")
def model():
    return main()


@pytest.mark.skipif(not helmholtz_available(), reason="General Helmholtz not available")
@pytest.mark.component
def test_basic_build(model):
    """Make a turbine model and make sure it doesn't throw exception"""
    # Check unit config arguments
    assert model.fs.Waterwalls[1].config.has_heat_transfer
    assert model.fs.Waterwalls[1].config.has_pressure_change
    assert model.fs.Waterwalls[1].config.property_package is model.fs.prop_water

    assert model.fs.downcomer.config.has_heat_transfer
    assert isinstance(model.fs.downcomer.heat_duty, pyo.Var)
    assert isinstance(model.fs.downcomer.deltaP, pyo.Var)
    assert isinstance(model.fs.drum.drum_level, pyo.Var)


# @pytest.mark.skipif(not helmholtz_available(), reason="General Helmholtz not available")
# @pytest.mark.integration
# def test_unit_consistency(model):
#     assert_units_consistent(model)


@pytest.mark.skipif(not helmholtz_available(), reason="General Helmholtz not available")
@pytest.mark.component
def test_init(model):
    # check that the model solved properly and has 0 degrees of freedom
    assert degrees_of_freedom(model) == 0

    model.fs.drum.feedwater_inlet.flow_mol[:].fix()
    model.fs.drum.feedwater_inlet.pressure[:].unfix()
    model.fs.drum.feedwater_inlet.enth_mol[:].fix()
    optarg = {"tol": 1e-6, "max_iter": 20}
    solver.options = optarg
    # set scaling parameters
    for i in model.fs.ww_zones:
        iscale.set_scaling_factor(model.fs.Waterwalls[i].heat_flux_conv[0], 1e-5)
    iscale.calculate_scaling_factors(model)
    results = solver.solve(model, tee=True)
    assert pyo.check_optimal_termination(results)

    assert pyo.value(model.fs.downcomer.deltaP[0]) > 0

    assert pytest.approx(0, abs=1e-3) == pyo.value(
        model.fs.Waterwalls[10].control_volume.properties_out[0].flow_mol
        - model.fs.Waterwalls[1].control_volume.properties_in[0].flow_mol
    )

    assert pyo.value(
        model.fs.drum.feedwater_inlet.flow_mol[0]
        - model.fs.drum.steam_outlet.flow_mol[0]
    ) == pytest.approx(0, abs=1e-3)



#################################################################################
# The Institute for the Design of Advanced Energy Systems Integrated Platform
# Framework (IDAES IP) was produced under the DOE Institute for the
# Design of Advanced Energy Systems (IDAES).
#
# Copyright (c) 2018-2024 by the software owners: The Regents of the
# University of California, through Lawrence Berkeley National Laboratory,
# National Technology & Engineering Solutions of Sandia, LLC, Carnegie Mellon
# University, West Virginia University Research Corporation, et al.
# All rights reserved.  Please see the files COPYRIGHT.md and LICENSE.md
# for full copyright and license information.
#################################################################################
"""Predefined expression for Helmholtz EoS functions
"""

__author__ = "John Eslick"

import pyomo.environ as pyo


def phi_residual_expressions_type01(model, parameters):
    """Type01 expression for the residual part of dimensionless Helmholtz free energy

    Args:
        model (Block): Pyomo model
        parameters (dict): Main parameters dictionary

    Returns:
        dict: Expressions for residual part of Helmholtz free energy
    """
    last_terms = parameters["eos"]["last_term_residual"]
    n = parameters["eos"]["n"]
    t = parameters["eos"]["t"]
    d = parameters["eos"]["d"]
    c = parameters["eos"]["c"]
    first_term = 1
    rng = []
    for last_term in last_terms:
        rng.append(range(first_term, last_term + 1))
        first_term = last_term + 1
    return {
        "phir": sum(n[i] * model.delta ** d[i] * model.tau ** t[i] for i in rng[0])
        + sum(
            n[i]
            * model.delta ** d[i]
            * model.tau ** t[i]
            * pyo.exp(-model.delta ** c[i])
            for i in rng[1]
        ),
        "phir_d": sum(
            n[i] * d[i] * model.delta ** (d[i] - 1) * model.tau ** t[i] for i in rng[0]
        )
        + sum(
            n[i]
            * pyo.exp(-model.delta ** c[i])
            * model.delta ** (d[i] - 1)
            * model.tau ** t[i]
            * (d[i] - c[i] * model.delta ** c[i])
            for i in rng[1]
        ),
        "phir_dd": sum(
            n[i] * d[i] * (d[i] - 1) * model.delta ** (d[i] - 2) * model.tau ** t[i]
            for i in rng[0]
        )
        + sum(
            n[i]
            * pyo.exp(-model.delta ** c[i])
            * model.delta ** (d[i] - 2)
            * model.tau ** t[i]
            * (
                (d[i] - c[i] * model.delta ** c[i])
                * (d[i] - 1 - c[i] * model.delta ** c[i])
                - c[i] ** 2 * model.delta ** c[i]
            )
            for i in rng[1]
        ),
        "phir_t": sum(
            n[i] * t[i] * model.delta ** d[i] * model.tau ** (t[i] - 1) for i in rng[0]
        )
        + sum(
            n[i]
            * t[i]
            * model.delta ** d[i]
            * model.tau ** (t[i] - 1)
            * pyo.exp(-model.delta ** c[i])
            for i in rng[1]
        ),
        "phir_tt": sum(
            n[i] * t[i] * (t[i] - 1) * model.delta ** d[i] * model.tau ** (t[i] - 2)
            for i in rng[0]
        )
        + sum(
            n[i]
            * t[i]
            * (t[i] - 1)
            * model.delta ** d[i]
            * model.tau ** (t[i] - 2)
            * pyo.exp(-model.delta ** c[i])
            for i in rng[1]
        ),
        "phir_dt": sum(
            n[i] * t[i] * d[i] * model.delta ** (d[i] - 1) * model.tau ** (t[i] - 1)
            for i in rng[0]
        )
        + sum(
            n[i]
            * t[i]
            * model.delta ** (d[i] - 1)
            * model.tau ** (t[i] - 1)
            * (d[i] - c[i] * model.delta ** c[i])
            * pyo.exp(-model.delta ** c[i])
            for i in rng[1]
        ),
    }



#################################################################################
# The Institute for the Design of Advanced Energy Systems Integrated Platform
# Framework (IDAES IP) was produced under the DOE Institute for the
# Design of Advanced Energy Systems (IDAES).
#
# Copyright (c) 2018-2024 by the software owners: The Regents of the
# University of California, through Lawrence Berkeley National Laboratory,
# National Technology & Engineering Solutions of Sandia, LLC, Carnegie Mellon
# University, West Virginia University Research Corporation, et al.
# All rights reserved.  Please see the files COPYRIGHT.md and LICENSE.md
# for full copyright and license information.
#################################################################################
# matopt.materials.parsers

# TODO: Decide if (and how) we should support:
# from matopt.materials.parsers import *
# NOTE: Probably will never need this, since parsers should
#       probably just be incorporated in Design.py and most
#       of these methods should be behind-the-scenes.

"""
import matopt.materials.parsers.CFG
import matopt.materials.parsers.PDB
import matopt.materials.parsers.POSCAR
import matopt.materials.parsers.XYZ
"""
# TODO: Decide if this is necessary
